import os
from tempfile import TemporaryDirectory

from pymatgen.io.lammps.outputs import (
    parse_lammps_dumps,
    parse_lammps_log,
)

from pyiron_atomistics.lammps.potential import (
    LammpsPotential,
    LammpsPotentialFile,
    list_potentials,
)
from pyiron_atomistics.lammps.control import LammpsControl

from pyiron_contrib.tinybase.container import (
    AbstractInput,
    AbstractOutput,
    StorageAttribute,
    StructureInput,
    EnergyPotOutput,
    EnergyKinOutput,
    ForceOutput,
)
from pyiron_contrib.tinybase.task import (
    AbstractTask,
    ReturnStatus,
)
from pyiron_contrib.tinybase.shell import ShellTask, ExecutablePathResolver


class LammpsInputInput(AbstractInput):
    working_directory = StorageAttribute().type(str)

    calc_type = StorageAttribute()

    def calc_static(self):
        self.calc_type = "static"

    def check_ready(self):
        return (
            self.working_directory is not None
            and self.calc_type == "static"
            and super().check_ready()
        )


class LammpsInputOutput(AbstractOutput):
    working_directory = StorageAttribute().type(str)


class LammpsInputTask(AbstractTask):
    """
    Write a set of input files for lammps calculations.

    The potential input is written to `potential.inp` together with any
    additional files needed by the potential.

    The structure is written to `structure.inp`.

    The lammps input script is to `control.inp`.
    """

    def _get_input(self):
        return LammpsInputInput()

    def _get_output(self):
        return LammpsInputOutput()

    def _execute(self, output):
        with open(
            os.path.join(self.input.working_directory, "structure.inp"), "w"
        ) as f:
            self.input.structure.write(f, format="lammps-data")

        potential = LammpsPotential()
        potential.df = LammpsPotentialFile().find_by_name(self.input.potential)
        potential.write_file(
            file_name="potential.inp", cwd=self.input.working_directory
        )
        potential.copy_pot_files(self.input.working_directory)

        control = LammpsControl()
        assert self.input.calc_type == "static", "Cannot happen"
        control.calc_static()
        control.write_file(file_name="control.inp", cwd=self.input.working_directory)

        output.working_directory = self.input.working_directory


class LammpsStaticParserInput(AbstractInput):
    working_directory = StorageAttribute().type(str)

    def check_ready(self):
        return self.working_directory is not None and super().check_ready()


class LammpsStaticOutput(EnergyPotOutput, EnergyKinOutput, ForceOutput):
    pass


class LammpsStaticParserTask(AbstractTask):
    """
    Parse a static lammps calculation.

    In practice return the last step of lammps dump and log file.
    """

    def _get_input(self):
        return LammpsStaticParserInput()

    def _get_output(self):
        return LammpsStaticOutput()

    def _execute(self, output):
        log = parse_lammps_log(
            os.path.join(self.input.working_directory, "log.lammps")
        )[-1]
        output.energy_pot = energy_pot = log["PotEng"].iloc[-1]
        output.energy_kin = log["TotEng"].iloc[-1] - energy_pot
        dump = list(
            parse_lammps_dumps(os.path.join(self.input.working_directory, "dump.out"))
        )[-1]
        output.forces = dump.data[["fx", "fy", "fz"]].to_numpy()


class LammpsInput(StructureInput):
    potential = StorageAttribute().type(str)

    def list_potentials(self):
        """
        List available potentials compatible with a set structure.

        If no structure is set, return an empty list.
        """
        if self.structure is not None:
            return list_potentials(self.structure)
        else:
            return []

    def check_ready(self):
        return self.potential is not None and super().check_ready()


class LammpsStaticTask(AbstractTask):
    """
    A static calculation with lammps.
    """

    def _get_input(self):
        return LammpsInput()

    def _get_output(self):
        return LammpsStaticOutput()

    def _execute(self, output):
        with TemporaryDirectory() as tmp_dir:
            inp = LammpsInputTask(capture_exceptions=self._capture_exceptions)
            inp.input.working_directory = tmp_dir
            inp.input.structure = self.input.structure
            inp.input.potential = self.input.potential
            inp.input.calc_static()
            ret, out = inp.execute()
            if not ret.is_done():
                return ReturnStatus.aborted(f"Writing input failed: {ret.msg}")

            lmp = ShellTask(capture_exceptions=self._capture_exceptions)
            lmp.input.command = ExecutablePathResolver("lammps", "lammps")
            lmp.input.working_directory = tmp_dir
            ret, out = lmp.execute()
            if not ret.is_done():
                return ReturnStatus.aborted(f"Running lammps failed: {ret.msg}")

            psr = LammpsStaticParserTask(capture_exceptions=self._capture_exceptions)
            psr.input.working_directory = tmp_dir
            ret, out = psr.execute()
            if not ret.is_done():
                return ReturnStatus.aborted(f"Parsing failed: {ret.msg}")
            output.take(out)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import os
from pyiron_atomistics.lammps.base import Input
from pyiron_atomistics.lammps.interactive import LammpsInteractive
from pyiron_atomistics.atomistics.structure.atoms import Atoms
from pyiron_atomistics.atomistics.structure.structurestorage import StructureStorage
from pyiron_contrib.atomistics.mlip.mlip import read_cgfs
from pyiron_contrib.atomistics.mlip.cfgs import loadcfgs
from pyiron_base import GenericParameters

import numpy as np

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"


class LammpsMlip(LammpsInteractive):
    def __init__(self, project, job_name):
        super(LammpsMlip, self).__init__(project, job_name)
        self.input = MlipInput()
        self.__name__ = "LammpsMlip"
        self.__version__ = None
        # Reset the version number to the executable is set automatically
        self._executable = None
        self._executable_activate()
        self._selected_structures = None

    def set_input_to_read_only(self):
        """
        This function enforces read-only mode for the input classes, but it has to be implement in the individual
        classes.
        """
        super(LammpsMlip, self).set_input_to_read_only()
        self.input.mlip.read_only = True

    def write_input(self):
        super(LammpsMlip, self).write_input()
        if self.input.mlip["mtp-filename"] == "auto":
            self.input.mlip["mtp-filename"] = os.path.basename(
                self.potential["Filename"][0][0]
            )
        self.input.mlip.write_file(file_name="mlip.ini", cwd=self.working_directory)

    def convergence_check(self):
        for line in self["error.out"]:
            if line.startswith("MLIP: Breaking threshold exceeded"):
                return False
        return True

    def enable_active_learning(self, threshold=2.0, threshold_break=5.0):
        """
        Enable active learning during MD run.

        Automatically collect structures on which the potential is extrapolating.

        Args:
            threshold (float): select structures with extrapolation grade larger than this
            threshold_break (float): stop the MD run after seeing a structure with extrapolation grade larger than this
        """
        self.executable.accepted_return_codes += [8]
        self.input.mlip.load_string(
            f"""\
mtp-filename auto
calculate-efs TRUE
select TRUE
select:threshold {threshold}
select:threshold-break {threshold_break}
select:save-selected selected.cfg
select:load-state state.mvs
select:log selection.log
write-cfgs:skip 0
"""
        )

    def _get_selection_file(self):
        return os.path.join(
            self.working_directory, self.input.mlip["select:save-selected"]
        )

    @property
    def _selection_enabled(self):
        return self.input.mlip["select"] == "TRUE"

    @property
    def selected_structures(self):
        """
        :class:`.StructureStorage`: structures that the potential extrapolated on during the run.

        Only available if :method:`.enable_active_learning` was called and once the job has been collected.
        """
        if not (
            self.status.collect or self.status.finished or self.status.not_converged
        ):
            raise ValueError(
                "Selected structures are only available once the job has finished!"
            )
        if not self._selection_enabled:
            raise ValueError(
                "Selected structures are only available after calling enable_active_learning()!"
            )
        if self._selected_structures is None:
            if "selected" in self["output"].list_groups():
                self._selected_structures = self["output/selected"].to_object()
            else:
                self._selected_structures = StructureStorage()
        return self._selected_structures

    def collect_output(self):
        super(LammpsMlip, self).collect_output()
        if "select:save-selected" in self.input.mlip._dataset["Parameter"]:
            file_name = self._get_selection_file()
            if os.path.exists(file_name):
                for cfg in loadcfgs(file_name):
                    species = np.array(self.potential.Species.iloc[0])[
                        cfg.types.astype(int)
                    ]
                    self.selected_structures.add_structure(
                        Atoms(
                            symbols=species,
                            positions=cfg.pos,
                            cell=cfg.lat,
                            pbc=[True, True, True],
                        ),
                        mv_grade=cfg.grade,
                    )
                (
                    cell,
                    positions,
                    forces,
                    stress,
                    energy,
                    indicies,
                    grades,
                    jobids,
                    timesteps,
                ) = read_cgfs(file_name=file_name)
                with self.project_hdf5.open("output/mlip") as hdf5_output:
                    hdf5_output["forces"] = forces
                    hdf5_output["energy_tot"] = energy
                    hdf5_output["pressures"] = stress
                    hdf5_output["cells"] = cell
                    hdf5_output["positions"] = positions
                    hdf5_output["indicies"] = indicies
            self.selected_structures.to_hdf(
                self.project_hdf5.open("output"), "selected"
            )


class MlipInput(Input):
    def __init__(self):
        self.mlip = MlipParameter()
        super(MlipInput, self).__init__()

    def to_hdf(self, hdf5):
        """

        Args:
            hdf5:
        Returns:
        """
        with hdf5.open("input") as hdf5_input:
            self.mlip.to_hdf(hdf5_input)
        super(MlipInput, self).to_hdf(hdf5)

    def from_hdf(self, hdf5):
        """

        Args:
            hdf5:
        Returns:
        """
        with hdf5.open("input") as hdf5_input:
            self.mlip.from_hdf(hdf5_input)
        super(MlipInput, self).from_hdf(hdf5)


class MlipParameter(GenericParameters):
    def __init__(self, separator_char=" ", comment_char="#", table_name="mlip_inp"):
        super(MlipParameter, self).__init__(
            separator_char=separator_char,
            comment_char=comment_char,
            table_name=table_name,
        )

    def load_default(self, file_content=None):
        if file_content is None:
            file_content = """\
mtp-filename auto
select FALSE
"""
        self.load_string(file_content)



"""
A base class for nodal objects that have internal structure -- i.e. they hold a
sub-graph
"""

from __future__ import annotations

from abc import ABC
from functools import partial
from typing import Optional
from warnings import warn

from pyiron_contrib.executors import CloudpickleProcessPoolExecutor
from pyiron_contrib.workflow.node import Node
from pyiron_contrib.workflow.function import (
    Function,
    SingleValue,
    Slow,
    function_node,
    slow_node,
    single_value_node,
)
from pyiron_contrib.workflow.node_library import atomistics, standard
from pyiron_contrib.workflow.node_library.package import NodePackage
from pyiron_contrib.workflow.util import DotDict


class _NodeDecoratorAccess:
    """An intermediate container to store node-creating decorators as class methods."""

    function_node = function_node
    slow_node = slow_node
    single_value_node = single_value_node


class Creator:
    """A shortcut interface for creating non-Node objects from the workflow class."""

    CloudpickleProcessPoolExecutor = CloudpickleProcessPoolExecutor


class Composite(Node, ABC):
    """
    A base class for nodes that have internal structure -- i.e. they hold a sub-graph.

    Item and attribute access is modified to give access to owned nodes.
    Adding a node with the `add` functionality or by direct attribute assignment sets
    this object as the parent of that node.

    Guarantees that each owned node is unique, and does not belong to any other parents.

    Offers a class method (`wrap_as`) to give easy access to the node-creating
    decorators.

    Specifies the required `on_run()` to call `run()` on a subset of owned nodes, i.e.
    to kick-start computation on the owned sub-graph.
    By default, `run()` will be called on all owned nodes have output connections but no
    input connections (i.e. the upstream-most nodes), but this can be overridden to
    specify particular nodes to use instead.
    The `run()` method (and `update()`, and calling the workflow, when these result in
    a run), return a new dot-accessible dictionary of keys and values created from the
    composite output IO panel.

    Does not specify `input` and `output` as demanded by the parent class; this
    requirement is still passed on to children.

    Attributes:
        nodes (DotDict[pyiron_contrib.workflow.node,Node]): The owned nodes that
         form the composite subgraph.
        strict_naming (bool): When true, repeated assignment of a new node to an
         existing node label will raise an error, otherwise the label gets appended
         with an index and the assignment proceeds. (Default is true: disallow assigning
         to existing labels.)
        add (NodeAdder): A tool for adding new nodes to this subgraph.
        upstream_nodes (list[pyiron_contrib.workflow.node,Node]): All the owned
         nodes that have output connections but no input connections, i.e. the
         upstream-most nodes.
        starting_nodes (None | list[pyiron_contrib.workflow.node,Node]): A subset
         of the owned nodes to be used on running. (Default is None, running falls back
         on using the `upstream_nodes`.)

    Methods:
        add(node: Node): Add the node instance to this subgraph.
        remove(node: Node): Break all connections the node has, remove it from this
         subgraph, and set its parent to `None`.
    """

    wrap_as = _NodeDecoratorAccess  # Class method access to decorators
    # Allows users/devs to easily create new nodes when using children of this class

    create = Creator

    def __init__(
        self,
        label: str,
        *args,
        parent: Optional[Composite] = None,
        run_on_updates: bool = True,
        strict_naming: bool = True,
        **kwargs,
    ):
        super().__init__(
            *args, label=label, parent=parent, run_on_updates=run_on_updates, **kwargs
        )
        self.strict_naming: bool = strict_naming
        self.nodes: DotDict[str:Node] = DotDict()
        self.add: NodeAdder = NodeAdder(self)
        self.starting_nodes: None | list[Node] = None

    @property
    def executor(self) -> None:
        return None

    @executor.setter
    def executor(self, new_executor):
        if new_executor is not None:
            raise NotImplementedError(
                "Running composite nodes with an executor is not yet supported"
            )

    def to_dict(self):
        return {
            "label": self.label,
            "nodes": {n.label: n.to_dict() for n in self.nodes.values()},
        }

    @property
    def upstream_nodes(self) -> list[Node]:
        return [
            node
            for node in self.nodes.values()
            if node.outputs.connected and not node.inputs.connected
        ]

    @property
    def on_run(self):
        return self.run_graph

    @staticmethod
    def run_graph(self):
        starting_nodes = (
            self.upstream_nodes if self.starting_nodes is None else self.starting_nodes
        )
        for node in starting_nodes:
            node.run()
        return DotDict(self.outputs.to_value_dict())

    @property
    def run_args(self) -> dict:
        return {"self": self}

    def add_node(self, node: Node, label: Optional[str] = None) -> None:
        """
        Assign a node to the parent. Optionally provide a new label for that node.

        Args:
            node (pyiron_contrib.workflow.node.Node): The node to add.
            label (Optional[str]): The label for this node.

        Raises:
            TypeError: If the
        """
        if not isinstance(node, Node):
            raise TypeError(
                f"Only new node instances may be added, but got {type(node)}."
            )
        self._ensure_node_has_no_other_parent(node)
        label = self._get_unique_label(node.label if label is None else label)
        self._ensure_node_is_not_duplicated(node, label)

        self.nodes[label] = node
        node.label = label
        node.parent = self
        return node

    def _get_unique_label(self, label):
        if label in self.__dir__():
            if isinstance(getattr(self, label), Node):
                if self.strict_naming:
                    raise AttributeError(
                        f"{label} is already the label for a node. Please remove it "
                        f"before assigning another node to this label."
                    )
                else:
                    label = self._add_suffix_to_label(label)
            else:
                raise AttributeError(
                    f"{label} is an attribute or method of the {self.__class__} class, "
                    f"and cannot be used as a node label."
                )
        return label

    def _add_suffix_to_label(self, label):
        i = 0
        new_label = label
        while new_label in self.nodes.keys():
            new_label = f"{label}{i}"
            i += 1
        if new_label != label:
            warn(
                f"{label} is already a node; appending an index to the "
                f"node label instead: {new_label}"
            )
        return new_label

    def _ensure_node_has_no_other_parent(self, node: Node):
        if node.parent is not None and node.parent is not self:
            raise ValueError(
                f"The node ({node.label}) already belongs to the parent "
                f"{node.parent.label}. Please remove it there before trying to "
                f"add it to this parent ({self.label})."
            )

    def _ensure_node_is_not_duplicated(self, node: Node, label: str):
        if (
            node.parent is self
            and label != node.label
            and self.nodes[node.label] is node
        ):
            warn(
                f"Reassigning the node {node.label} to the label {label} when "
                f"adding it to the parent {self.label}."
            )
            del self.nodes[node.label]

    def remove(self, node: Node | str):
        if isinstance(node, Node):
            node.parent = None
            node.disconnect()
            del self.nodes[node.label]
        else:
            del self.nodes[node]

    def __setattr__(self, label: str, node: Node):
        if isinstance(node, Node):
            self.add_node(node, label=label)
        else:
            super().__setattr__(label, node)

    def __getattr__(self, key):
        return self.nodes[key]

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __iter__(self):
        return self.nodes.values().__iter__()

    def __len__(self):
        return len(self.nodes)

    def __dir__(self):
        return set(super().__dir__() + list(self.nodes.keys()))


class NodeAdder:
    """
    This class provides a layer of misdirection so that `Composite` objects can set
    themselves as the parent of owned nodes.

    It also provides access to packages of nodes and the ability to register new
    packages.
    """

    def __init__(self, parent: Composite):
        self._parent: Composite = parent
        self.register_nodes("atomistics", *atomistics.nodes)
        self.register_nodes("standard", *standard.nodes)

    Function = Function
    Slow = Slow
    SingleValue = SingleValue

    def __getattribute__(self, key):
        value = super().__getattribute__(key)
        if value == Function:
            return partial(Function, parent=self._parent)
        return value

    def __call__(self, node: Node):
        return self._parent.add_node(node)

    def register_nodes(self, domain: str, *nodes: list[type[Node]]):
        """
        Add a list of node classes to be accessible for creation under the provided
        domain name.

        TODO: multiple dispatch so we can handle registering something other than a
              list, e.g. modules or even urls.
        """
        setattr(self, domain, NodePackage(self._parent, *nodes))



from __future__ import annotations

from typing import Optional

from pyiron_atomistics import Project, _StructureFactory
from pyiron_atomistics.atomistics.job.atomistic import AtomisticGenericJob
from pyiron_atomistics.atomistics.structure.atoms import Atoms
from pyiron_atomistics.lammps.lammps import Lammps as LammpsJob

from pyiron_contrib.workflow.function import single_value_node, slow_node


@single_value_node(output_labels="structure")
def bulk_structure(element: str = "Fe", cubic: bool = False, repeat: int = 1) -> Atoms:
    return _StructureFactory().bulk(element, cubic=cubic).repeat(repeat)


@single_value_node(output_labels="job")
def lammps(structure: Optional[Atoms] = None) -> LammpsJob:
    pr = Project(".")
    job = pr.atomistics.job.Lammps("NOTAREALNAME")
    job.structure = structure if structure is not None else _StructureFactory().bulk()
    job.potential = job.list_potentials()[0]
    return job


def _run_and_remove_job(job, modifier: Optional[callable] = None, **modifier_kwargs):
    """
    Extracts the commonalities for all the "calc" methods for running a Lammps engine.
    Will need to be extended/updated once we support other engines so that more output
    can be parsed. Output may wind up more concretely packaged, e.g. as `CalcOutput` or
    `MDOutput`, etc., ala Joerg's suggestion later, so for the time being we don't put
    too much effort into this.

    Warning:
        Jobs are created in a dummy project with a dummy name and are all removed at the
        end; this works fine for serial workflows, but will need to be revisited --
        probably with naming based on the parantage of node/workflow labels -- once
        other non-serial execution is introduced.
    """
    job_name = "JUSTAJOBNAME"
    pr = Project("WORKFLOWNAMEPROJECT")
    job = job.copy_to(project=pr, new_job_name=job_name, delete_existing_job=True)
    if modifier is not None:
        job = modifier(job, **modifier_kwargs)
    job.run()

    cells = job.output.cells
    displacements = job.output.displacements
    energy_pot = job.output.energy_pot
    energy_tot = job.output.energy_tot
    force_max = job.output.force_max
    forces = job.output.forces
    indices = job.output.indices
    positions = job.output.positions
    pressures = job.output.pressures
    steps = job.output.steps
    temperature = job.output.temperature
    total_displacements = job.output.total_displacements
    unwrapped_positions = job.output.unwrapped_positions
    volume = job.output.volume

    job.remove()
    pr.remove(enable=True)

    return (
        cells,
        displacements,
        energy_pot,
        energy_tot,
        force_max,
        forces,
        indices,
        positions,
        pressures,
        steps,
        temperature,
        total_displacements,
        unwrapped_positions,
        volume,
    )


@slow_node(
    output_labels=[
        "cells",
        "displacements",
        "energy_pot",
        "energy_tot",
        "force_max",
        "forces",
        "indices",
        "positions",
        "pressures",
        "steps",
        "temperature",
        "total_displacements",
        "unwrapped_positions",
        "volume",
    ]
)
def calc_static(
    job: AtomisticGenericJob,
):
    return _run_and_remove_job(job=job)


@slow_node(
    output_labels=[
        "cells",
        "displacements",
        "energy_pot",
        "energy_tot",
        "force_max",
        "forces",
        "indices",
        "positions",
        "pressures",
        "steps",
        "temperature",
        "total_displacements",
        "unwrapped_positions",
        "volume",
    ]
)
def calc_md(
    job: AtomisticGenericJob,
    n_ionic_steps: int = 1000,
    n_print: int = 100,
    temperature: int | float = 300.0,
    pressure: float
    | tuple[float, float, float]
    | tuple[float, float, float, float, float, float]
    | None = None,
):
    def calc_md(job, n_ionic_steps, n_print, temperature, pressure):
        job.calc_md(
            n_ionic_steps=n_ionic_steps,
            n_print=n_print,
            temperature=temperature,
            pressure=pressure,
        )
        return job

    return _run_and_remove_job(
        job=job,
        modifier=calc_md,
        n_ionic_steps=n_ionic_steps,
        n_print=n_print,
        temperature=temperature,
        pressure=pressure,
    )


nodes = [
    bulk_structure,
    calc_md,
    calc_static,
    lammps,
]



"""
Channels are access points for information to flow into and out of nodes.

Data channels carry, unsurprisingly, data.
Input data channels force an update on their owning node when they are updated with new
data, and output data channels update all the input data channels to which they are
connected.
In this way, data channels facilitate forward propagation of data through a graph.
They hold data persistently.

Signal channels are tools for procedurally exposing functionality on nodes.
Input signal channels are connected to a callback function which gets invoked when the
channel is updated.
Output signal channels must be accessed by the owning node directly, and then trigger
all the input signal channels to which they are connected.
In this way, signal channels can force behaviour (node method calls) to propagate
forwards through a graph.
They do not hold any data, but rather fire for an effect.
"""

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from warnings import warn

from pyiron_contrib.workflow.has_channel import HasChannel
from pyiron_contrib.workflow.has_to_dict import HasToDict
from pyiron_contrib.workflow.type_hinting import (
    valid_value,
    type_hint_is_as_or_more_specific_than,
)

if typing.TYPE_CHECKING:
    from pyiron_contrib.workflow.node import Node


class Channel(HasChannel, HasToDict, ABC):
    """
    Channels facilitate the flow of information (data or control signals) into and
    out of nodes.
    They must have a label and belong to a node.

    Input/output channels can be (dis)connected from other output/input channels, and
    store all of their current connections in a list.
    This connection information is duplicated in that it is stored on _both_ channels
    that form the connection.

    Child classes must define a string representation, `__str__`, and what to do on an
    attempted connection, `connect`.

    Attributes:
        label (str): The name of the channel.
        node (pyiron_contrib.workflow.node.Node): The node to which the channel
         belongs.
        connections (list[Channel]): Other channels to which this channel is connected.
    """

    def __init__(
        self,
        label: str,
        node: Node,
    ):
        """
        Make a new channel.

        Args:
            label (str): A name for the channel.
            node (pyiron_contrib.workflow.node.Node): The node to which the
             channel belongs.
        """
        self.label: str = label
        self.node: Node = node
        self.connections: list[Channel] = []

    @abstractmethod
    def __str__(self):
        pass

    @abstractmethod
    def connect(self, *others: Channel) -> None:
        """
        How to handle connections to other channels.

        Args:
            *others (Channel): The other channel objects to attempt to connect with.
        """
        pass

    def disconnect(self, *others: Channel) -> None:
        """
        If currently connected to any others, removes this and the other from eachothers
        respective connections lists.

        Args:
            *others (Channel): The other channels to disconnect from.
        """
        for other in others:
            if other in self.connections:
                self.connections.remove(other)
                other.disconnect(self)

    def disconnect_all(self) -> None:
        """
        Disconnect from all other channels currently in the connections list.
        """
        self.disconnect(*self.connections)

    @property
    def connected(self) -> bool:
        """
        Has at least one connection.
        """
        return len(self.connections) > 0

    def _already_connected(self, other: Channel) -> bool:
        return other in self.connections

    def __iter__(self):
        return self.connections.__iter__()

    def __len__(self):
        return len(self.connections)

    @property
    def channel(self) -> Channel:
        return self

    def to_dict(self) -> dict:
        return {
            "label": self.label,
            "connected": self.connected,
            "connections": [f"{c.node.label}.{c.label}" for c in self.connections],
        }


class NotData:
    """
    This class exists purely to initialize data channel values where no default value
    is provided; it lets the channel know that it has _no data in it_ and thus should
    not identify as ready.
    """

    @classmethod
    def __repr__(cls):
        # We use the class directly (not instances of it) where there is not yet data
        # So give it a decent repr, even as just a class
        return cls.__name__


class DataChannel(Channel, ABC):
    """
    Data channels control the flow of data on the graph.
    They store this data in a `value` attribute.
    They may optionally have a type hint.
    They have a `ready` attribute which tells whether their value matches their type
    hint (if one is provided, else `True`).
    They may optionally have a storage priority (but this doesn't do anything yet).
    (In the future they may optionally have an ontological type.)

    The `value` held by a channel can be manually assigned, but should normally be set
    by the `update` method.
    In neither case is the type hint strictly enforced.

    Type hinting is strictly enforced in one situation: when making connections to
    other channels and at least one data channel has a non-None value for its type hint.
    In this case, we insist that the output type hint be _as or more more specific_ than
    the input type hint, to ensure that the input always receives output of a type it
    expects. This behaviour can be disabled and all connections allowed by setting
    `strict_connections = False` on the relevant input channel.

    For simple type hints like `int` or `str`, type hint comparison is trivial.
    However, some hints take arguments, e.g. `dict[str, int]` to specify key and value
    types; `tuple[int, int, str]` to specify a tuple with certain values;
    `typing.Literal['a', 'b', 'c']` to specify particular choices;
    `typing.Callable[[float, float], str]` to specify a callable that takes particular
    argument types and has a return type; etc.
    For hints with the origin `dict`, `tuple`, and `typing.Callable`, the two hints must
    have _exactly the same arguments_ for one two qualify as "as or more specific".
    E.g. `tuple[int, int|float]` is as or more specific than
    `tuple[int|float, int|float]`, but not `tuple[int, int|float, str]`.
    For _all other hints_, we demand that the output hint arguments be a _subset_ of
    the input.
    E.g. `Literal[1, 2]` is as or more specific that both `Literal[1, 2]` and
    `Literal[1, 2, "three"]`.

    The data `value` will initialize to an instance of `NotData` by default.
    The channel will identify as `ready` when the value is _not_ an instance of
    `NotData`, and when the value conforms to type hints (if any).

    Warning:
        Type hinting in python is quite complex, and determining when a hint is
        "more specific" can be tricky. For instance, in python 3.11 you can now type
        hint a tuple with a mixture of fixed elements of fixed type, followed by an
        arbitrary elements of arbitrary type. This and other complex scenarios are not
        yet included in our test suite and behaviour is not guaranteed.
    """

    def __init__(
        self,
        label: str,
        node: Node,
        default: typing.Optional[typing.Any] = NotData,
        type_hint: typing.Optional[typing.Any] = None,
    ):
        super().__init__(label=label, node=node)
        self.default = default
        self.value = default
        self.type_hint = type_hint

    @property
    def ready(self) -> bool:
        """
        Check if the currently stored value satisfies the channel's type hint.

        Returns:
            (bool): Whether the value matches the type hint.
        """
        if self.type_hint is not None:
            return self._value_is_data and valid_value(self.value, self.type_hint)
        else:
            return self._value_is_data

    @property
    def _value_is_data(self):
        return self.value is not NotData

    def update(self, value) -> None:
        """
        Store a new value and trigger before- and after-update routines.

        Args:
            value: The value to store.
        """
        self._before_update()
        self.value = value
        self._after_update()

    def _before_update(self) -> None:
        """
        A tool for child classes to do things before the value changed during an update.
        """
        pass

    def _after_update(self) -> None:
        """
        A tool for child classes to do things after the value changed during an update.
        """
        pass

    def connect(self, *others: DataChannel) -> None:
        """
        For all others for which the connection is valid (one input, one output, both
        data channels), adds this to the other's list of connections and the other to
        this list of connections.
        Then the input channel gets updated with the output channel's current value.

        Args:
            *others (DataChannel):

        Raises:
            TypeError: When one of others is not a `DataChannel`
        """
        for other in others:
            if self._valid_connection(other):
                self.connections.append(other)
                other.connections.append(self)
                out, inp = self._figure_out_who_is_who(other)
                inp.update(out.value)
            else:
                if isinstance(other, DataChannel):
                    warn(
                        f"{self.label} ({self.__class__.__name__}) and {other.label} "
                        f"({other.__class__.__name__}) were not a valid connection"
                    )
                else:
                    raise TypeError(
                        f"Can only connect two channels, but {self.label} "
                        f"({self.__class__.__name__}) got a {other} ({type(other)})"
                    )

    def _valid_connection(self, other) -> bool:
        if self._is_IO_pair(other) and not self._already_connected(other):
            if self._both_typed(other):
                out, inp = self._figure_out_who_is_who(other)
                if not inp.strict_connections:
                    return True
                else:
                    return type_hint_is_as_or_more_specific_than(
                        out.type_hint, inp.type_hint
                    )
            else:
                # If either is untyped, don't do type checking
                return True
        else:
            return False

    def _is_IO_pair(self, other: DataChannel) -> bool:
        return isinstance(other, DataChannel) and not isinstance(other, self.__class__)

    def _both_typed(self, other: DataChannel) -> bool:
        return self.type_hint is not None and other.type_hint is not None

    def _figure_out_who_is_who(self, other: DataChannel) -> (OutputData, InputData):
        return (self, other) if isinstance(self, OutputData) else (other, self)

    def __str__(self):
        return str(self.value)

    def to_dict(self) -> dict:
        d = super().to_dict()
        d["value"] = repr(self.value)
        d["ready"] = self.ready
        return d


class InputData(DataChannel):
    """
    On `update`, Input channels will then propagate their value along to their owning
    node by invoking its `update` method.

    `InputData` channels may be set to `wait_for_update()`, and they are only `ready`
    when they are not `waiting_for_update`.
    Their parent node can be told to always set them to wait for an update after the
    node runs using `require_update_after_node_runs()`.
    This allows nodes to complete the update of multiple channels before running again.

    The `strict_connections` parameter controls whether connections are subject to
    type checking requirements.
    I.e., they may set `strict_connections` to `False` (`True` -- default) at
    instantiation or later with `(de)activate_strict_connections()` to prevent (enable)
    data type checking when making connections with `OutputData` channels.
    """

    def __init__(
        self,
        label: str,
        node: Node,
        default: typing.Optional[typing.Any] = NotData,
        type_hint: typing.Optional[typing.Any] = None,
        strict_connections: bool = True,
    ):
        super().__init__(
            label=label,
            node=node,
            default=default,
            type_hint=type_hint,
        )
        self.strict_connections = strict_connections
        self.waiting_for_update = False

    def wait_for_update(self) -> None:
        """
        Sets `waiting_for_update` to `True`, which prevents `ready` from returning
        `True` until `update` is called.
        """
        self.waiting_for_update = True

    @property
    def ready(self) -> bool:
        """
        Extends the parent class check for whether the value matches the type hint with
        a check for whether the channel has been told to wait for an update (and not
        been updated since then).

        Returns:
            (bool): True when the stored value matches the type hint and the channel
                has not been told to wait for an update.
        """
        return not self.waiting_for_update and super().ready

    def _before_update(self) -> None:
        if self.node.running:
            raise RuntimeError(
                f"Parent node {self.node.label} of {self.label} is running, so value "
                f"cannot be updated."
            )

    def _after_update(self) -> None:
        self.waiting_for_update = False
        self.node.update()

    def require_update_after_node_runs(self, wait_now=False) -> None:
        """
        Registers this channel with its owning node as one that should have
        `wait_for_update()` applied after each time the node runs.

        Args:
            wait_now (bool): Also call `wait_for_update()` right now. (Default is
                False.)
        """
        if self.label not in self.node.channels_requiring_update_after_run:
            self.node.channels_requiring_update_after_run.append(self.label)
        if wait_now:
            self.wait_for_update()

    def activate_strict_connections(self) -> None:
        self.strict_connections = True

    def deactivate_strict_connections(self) -> None:
        self.strict_connections = False


class OutputData(DataChannel):
    """
    On `update`, Output channels propagate their value to all the input channels to
    which they are connected by invoking their `update` method.
    """

    def _after_update(self) -> None:
        for inp in self.connections:
            inp.update(self.value)


class SignalChannel(Channel, ABC):
    """
    Signal channels give the option control execution flow by triggering callback
    functions.

    Output channels can be called to trigger the callback functions of all input
    channels to which they are connected.
    """

    @abstractmethod
    def __call__(self) -> None:
        pass

    def connect(self, *others: SignalChannel) -> None:
        """
        For all others for which the connection is valid (one input, one output, both
        data channels), adds this to the other's list of connections and the other to
        this list of connections.

        Args:
            *others (SignalChannel): The other channels to attempt a connection to

        Raises:
            TypeError: When one of others is not a `SignalChannel`
        """
        for other in others:
            if self._valid_connection(other):
                self.connections.append(other)
                other.connections.append(self)
            else:
                if isinstance(other, SignalChannel):
                    warn(
                        f"{self.label} ({self.__class__.__name__}) and {other.label} "
                        f"({other.__class__.__name__}) were not a valid connection"
                    )
                else:
                    raise TypeError(
                        f"Can only connect two signal channels, but {self.label} "
                        f"({self.__class__.__name__}) got a {other} ({type(other)})"
                    )

    def _valid_connection(self, other) -> bool:
        return self._is_IO_pair(other) and not self._already_connected(other)

    def _is_IO_pair(self, other) -> bool:
        return isinstance(other, SignalChannel) and not isinstance(
            other, self.__class__
        )


class InputSignal(SignalChannel):
    """
    Invokes a callback when called.
    """

    def __init__(
        self,
        label: str,
        node: Node,
        callback: callable,
    ):
        """
        Make a new input signal channel.

        Args:
            label (str): A name for the channel.
            node (pyiron_contrib.workflow.node.Node): The node to which the
             channel belongs.
            callback (callable): An argument-free callback to invoke when calling this
                object.
        """
        super().__init__(label=label, node=node)
        self.callback: callable = callback

    def __call__(self) -> None:
        self.callback()

    def __str__(self):
        return f"{self.label} runs {self.callback.__name__}"

    def to_dict(self) -> dict:
        d = super().to_dict()
        d["callback"] = self.callback.__name__
        return d


class OutputSignal(SignalChannel):
    """
    Calls all the input signal objects in its connections list when called.
    """

    def __call__(self) -> None:
        for c in self.connections:
            c()

    def __str__(self):
        return (
            f"{self.label} activates "
            f"{[f'{c.node.label}.{c.label}' for c in self.connections]}"
        )



import numpy as np
import pint
from sklearn.linear_model import LinearRegression
from pyiron_atomistics.atomistics.master.parallel import AtomisticParallelMaster
from pyiron_base import JobGenerator


unit = pint.UnitRegistry()


def get_helmholtz_free_energy(frequencies, temperature, E_0=0):
    hn = (frequencies * 1e12 * unit.hertz * unit.planck_constant).to("eV").magnitude
    hn = hn[hn > 0]
    kBT = (temperature * unit.kelvin * unit.boltzmann_constant).to("eV").magnitude
    kBT = np.atleast_1d(kBT)
    values = np.zeros_like(kBT)
    values[kBT > 0] = kBT[kBT > 0] * np.log(
        1 - np.exp(-np.einsum("i,...->...i", hn, 1 / kBT[kBT > 0]))
    ).sum(axis=-1)
    return 0.5 * np.sum(hn) + values + E_0


def get_potential_energy(frequencies, temperature, E_0=0):
    hn = (frequencies * 1e12 * unit.hertz * unit.planck_constant).to("eV").magnitude
    kBT = (temperature * unit.kelvin * unit.boltzmann_constant).to("eV").magnitude
    kBT = np.atleast_1d(kBT)
    values = np.zeros_like(kBT)
    values[kBT > 0] = np.sum(
        hn / (np.exp(-np.einsum("i,...->...i", hn, 1 / kBT[kBT > 0])) - 1), axis=-1
    )
    return 0.5 * np.sum(hn) + values + E_0


def get_vibrational_frequencies(structure, hessian):
    m = np.tile(structure.get_masses(), (3, 1)).T.flatten()
    mass_tensor = np.sqrt(m * m[:, np.newaxis])
    nu_square = (
        (
            np.linalg.eigh(hessian / mass_tensor)[0]
            * unit.electron_volt
            / unit.angstrom**2
            / unit.amu
        )
        .to("THz**2")
        .magnitude
    )
    return np.sign(nu_square) * np.sqrt(np.absolute(nu_square)) / (2 * np.pi)


class Displacements:
    """Creation of displacement field."""

    def __init__(self, structure, symprec=1.0e-2, include_zero=True, second_order=True):
        """
        Displacement field

        Args:
            structure (pyiron_atomistics.atomistics.structure.atoms): pyiron structure
            symprec (float): Symmetry search precision
            include_zero (bool): Whether or not to include the non-diplaced structure. By
                including this, results get stabilized, while the computational cost does not
                increase significantly.
            second_order (bool): Whether or not to displace in both positive and negative
                direction. For symmetric displacement it is not performed.
        """
        self.structure = structure
        self.symprec = symprec
        self.include_zero = include_zero
        self.second_order = second_order
        self._sym = None
        self._unique_entries = None

    @property
    def symmetry(self):
        if self._sym is None:
            self._sym = self.structure.get_symmetry(symprec=self.symprec)
        return self._sym

    @property
    def unique_entries(self):
        if self._unique_entries is None:
            self._unique_entries = np.unique(
                self.symmetry.arg_equivalent_vectors, return_index=True
            )[1]
        return self._unique_entries

    def unravel(self, entries):
        return np.unravel_index(entries, self.structure.positions.shape)

    @property
    def even_entries(self):
        rand = np.random.random(self.structure.positions.shape)
        rand_sym = self.symmetry.symmetrize_vectors(rand)
        return ~np.isclose(rand_sym, 0)

    def get_unique_indices(self, positive=True):
        if positive:
            return self.unravel(self.unique_entries)
        return self.unravel(
            self.unique_entries[self.even_entries.flatten()[self.unique_entries]]
        )

    def get_displacements(self, positive=True):
        ind_x, ind_y = self.get_unique_indices(positive=positive)
        displacements = np.zeros((len(ind_x),) + self.structure.positions.shape)
        displacements[np.arange(len(ind_x)), ind_x, ind_y] = 1
        return displacements

    @property
    def displacements(self):
        d = self.get_displacements(positive=True)
        if self.include_zero:
            d = np.append([np.zeros_like(self.structure.positions)], d, axis=0)
        if self.second_order:
            d = np.append(d, self.get_displacements(positive=False), axis=0)
        return d


def generate_displacements(structure, symprec=1.0e-2):
    sym = structure.get_symmetry(symprec=symprec)
    _, comp = np.unique(sym.arg_equivalent_vectors, return_index=True)
    ind_x, ind_y = np.unravel_index(comp, structure.positions.shape)
    displacements = np.zeros((len(ind_x),) + structure.positions.shape)
    displacements[np.arange(len(ind_x)), ind_x, ind_y] = 1
    return displacements


thermo_doc = """
Get OUTPUT that minimizes the free energy.

Args:
    INPUT_ONE (float/ndarray): INPUT_ONE value(-s)
    INPUT_TWO (float/ndarray): INPUT_TWO value(-s)

Returns:
    OUTPUT (float/ndarray): OUTPUT value(-s)

Input values must be either both float, both ndarray of the same shape, or one of them
ndarray and the other one float.
"""


class Hessian:
    def __init__(
        self,
        structure,
        dx=0.01,
        symprec=1.0e-2,
        include_zero_displacement=True,
        second_order=True,
    ):
        """
        Calculation of the Hessian matrix

        Args:
            structure (pyiron_atomistics.structure.atoms.Atoms): Structure
            dx (float): Displacement (in distance unit)
            symprec (float): Symmetry search precision
            include_zero_displacement (bool): Whether to include zero
                displacement, which is not required to get the force constants,
                but increases the precision, especially if the initial
                positions are not the relaxed structure.
        """
        self.structure = structure.copy()
        self._symmetry = None
        self._permutations = None
        self.dx = dx
        self._displacements = []
        self._forces = None
        self._inequivalent_displacements = None
        self._symprec = symprec
        self._nu = None
        self._energy = None
        self._hessian = None
        self.include_zero_displacement = include_zero_displacement
        self.second_order = second_order

    @property
    def symmetry(self):
        if self._symmetry is None:
            self._symmetry = self.structure.get_symmetry(symprec=self._symprec)
        return self._symmetry

    def reset(self):
        self._inequivalent_displacements = None
        self._nu = None
        self._hessian = None

    @property
    def energy(self):
        return self._energy

    @energy.setter
    def energy(self, energy):
        if energy is not None and len(energy) != len(self.displacements):
            raise AssertionError(
                "Energy shape does not match existing displacement shape"
            )
        self._energy = np.array(energy)
        self.reset()

    @property
    def forces(self):
        return self._forces

    @forces.setter
    def forces(self, forces):
        if np.array(forces).shape != self.displacements.shape:
            raise AssertionError(
                "Force shape does not match existing displacement shape"
            )
        self._forces = np.array(forces)
        self.reset()

    @property
    def all_displacements(self):
        return np.einsum(
            "nxy,lnmy->lnmx",
            self.symmetry.rotations,
            self.displacements[:, self.symmetry.permutations],
        ).reshape(-1, np.prod(self.structure.positions.shape))

    @property
    def inequivalent_indices(self):
        return np.unique(self.all_displacements, axis=0, return_index=True)[1]

    @property
    def inequivalent_displacements(self):
        if self._inequivalent_displacements is None:
            self._inequivalent_displacements = self.all_displacements[
                self.inequivalent_indices
            ]
            self._inequivalent_displacements -= self.origin.flatten()
        return self._inequivalent_displacements

    @property
    def inequivalent_forces(self):
        forces = np.einsum(
            "nxy,lnmy->lnmx",
            self.symmetry.rotations,
            self.forces[:, self.symmetry.permutations],
        ).reshape(-1, np.prod(self.structure.positions.shape))
        return forces[self.inequivalent_indices]

    @property
    def _x_outer(self):
        return np.einsum("ik,ij->kj", *2 * [self.inequivalent_displacements])

    @property
    def hessian(self):
        if self._hessian is None:
            H = -np.einsum(
                "kj,in,ik->nj",
                np.linalg.inv(self._x_outer),
                self.inequivalent_forces,
                self.inequivalent_displacements,
                optimize=True,
            )
            self._hessian = 0.5 * (H + H.T)
        return self._hessian

    @property
    def vibrational_frequencies(self):
        if self._nu is None:
            self._nu = get_vibrational_frequencies(self.structure, self.hessian)
        return self._nu

    @property
    def minimum_displacements(self):
        return (
            Displacements(
                structure=self.structure,
                include_zero=self.include_zero_displacement,
                second_order=self.second_order,
                symprec=self._symprec,
            ).displacements
            * self.dx
        )

    @property
    def displacements(self):
        if len(self._displacements) == 0:
            self.displacements = self.minimum_displacements
        return np.asarray(self._displacements)

    @displacements.setter
    def displacements(self, d):
        self._displacements = np.asarray(d).tolist()
        self._inequivalent_displacements = None

    @property
    def _fit(self):
        E = self.energy + 0.5 * np.einsum("nij,nij->n", self.forces, self.displacements)
        E = np.repeat(E, len(self.symmetry.rotations))[self.inequivalent_indices]
        reg = LinearRegression()
        reg.fit(self.inequivalent_forces, E)
        return reg

    @property
    def min_energy(self):
        return self._fit.intercept_

    @property
    def origin(self):
        if self.energy is None or self.forces is None:
            return np.zeros_like(self.structure.positions)
        return 2 * self.symmetry.symmetrize_vectors(self._fit.coef_.reshape(-1, 3))

    @property
    def volume(self):
        return self.structure.get_volume()


class QHAJobGenerator(JobGenerator):
    @property
    def parameter_list(self):
        """

        Returns:
            (list)
        """
        cell, positions = self._master.get_supercells_with_displacements()
        return [[c, p] for c, p in zip(cell, positions)]

    @staticmethod
    def modify_job(job, parameter):
        job.structure.set_cell(parameter[0], scale_atoms=True)
        job.structure.set_scaled_positions(parameter[1])
        return job


class QuasiHarmonicApproximation(AtomisticParallelMaster):
    """
    Launch Quasi Harmonic Approximation jobs to calculate free energies and
    derived thermodynamic quantities.


    Example:

    >>> qha = pr.create.job.QuasiHarmonicApproximation('qha')
    >>> qha.ref_job = your_job
    >>> qha.run()
    >>> qha.get_helmholtz_free_energy(temperature=300, strain=0)


    Important differences to the existing `QuasiHarmonicJob`:

    - It does not do an internal multiplication of the structure, i.e. the
        input structure directly defines the interaction range. For this
        reason, you might be interested to make sure that the structure is
        large enough.
    - The minimum energy of each of the strained structures is estimated using
        the formula:

        .. math::

            U = U_0 + 0.5 * \sum_i forces_i * (dx_i - dX_i)

        where :math:`dx_i` are the internally generated displacement field,
        :math:`U` is the potential energy, :math:`forces_i` are the forces and
        :math:`U_0` and :math:`dX_i` are fitting parameters. The same formula
        also estimates the correct force-free positions and therefore the
        correct displacement field for the given strain.
    - If you don't need a volume or pressure dependence, you can set
        `self.input['num_points'] = 1`, in which case all values are returned
        for the given volume of the reference structure.
    """

    def __init__(self, project, job_name):
        """

        Args:
            project: project
            job_name: job name
        """
        super().__init__(project, job_name)
        self.__name__ = "QuasiHarmonicApproximation"
        self.__version__ = "0.0.1"
        self.input["displacement"] = (0.01, "Atom displacement magnitude")
        self.input["symprec"] = 1.0e-2
        self.input["include_zero_displacement"] = True
        self.input["second_order"] = True
        self.input["num_points"] = (11, "number of sample points")
        self.input["vol_range"] = (
            0.1,
            "relative volume variation around volume defined by ref_job",
        )
        self._job_generator = QHAJobGenerator(self)
        self._hessian = None
        self._td = None

    @property
    def strain_lst(self):
        if self.input["num_points"] == 1:
            return np.array([0])
        return (
            np.linspace(-1, 1, self.input["num_points"]) * self.input["vol_range"] / 2
        )

    @property
    def hessian(self):
        if self._hessian is None:
            self._hessian = Hessian(
                self.structure,
                include_zero_displacement=self.input["include_zero_displacement"],
                second_order=self.input["second_order"],
                symprec=self.input["symprec"],
            )
        return self._hessian

    def get_supercells_with_displacements(self):
        dx = self.structure.positions + self.hessian.displacements
        sc_dx = np.einsum("ij,...i->...j", np.linalg.inv(self.structure.cell), dx)
        positions = np.concatenate((len(self.strain_lst) * [sc_dx]))
        cell = np.repeat(
            self.structure.cell * (1 + self.strain_lst[:, np.newaxis, np.newaxis]),
            len(self.hessian.displacements),
            axis=0,
        )
        return cell, positions

    def to_hdf(self, hdf=None, group_name=None):
        """
        Store the QHA in an HDF5 file

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super().to_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("input") as hdf5_input:
            hdf5_input["displacements"] = self.hessian.displacements

    def from_hdf(self, hdf=None, group_name=None):
        """
        Restore the QHA from an HDF5 file

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super().from_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("input") as hdf5_input:
            if "displacements" in hdf5_input.list_nodes():
                self.hessian.displacements = hdf5_input["displacements"]

    def _load_output(self):
        if self.ref_job.server.run_mode.interactive:
            inspect = self.project_hdf5.inspect(self.child_ids[0])
            forces_lst = inspect["output/generic/forces"]
            energy_lst = inspect["output/generic/energy_pot"]
        else:
            pr_job = self.project_hdf5.project.open(self.job_name + "_hdf5")
            forces_lst, energy_lst = [], []
            for job_name in self.child_names:
                inspect = pr_job.inspect(job_name)
                forces_lst.append(inspect["output/generic/forces"])
                energy_lst.append(inspect["output/generic/energy_pot"])
        forces_lst = np.asarray(forces_lst).reshape(
            (self.input["num_points"],) + self.hessian.displacements.shape
        )
        energy_lst = np.asarray(energy_lst).reshape(
            self.input["num_points"], len(self.hessian.displacements)
        )
        return forces_lst, energy_lst

    def collect_output(self):
        """

        Returns:

        """
        forces, energy = self._load_output()
        nu_lst, h_lst, e_lst = [], [], []
        for f, e in zip(forces, energy):
            self.hessian.forces = f
            self.hessian.energy = e
            h_lst.append(self.hessian.hessian)
            nu_lst.append(self.hessian.vibrational_frequencies)
            e_lst.append(self.hessian.min_energy)

        with self.project_hdf5.open("output") as hdf5_out:
            hdf5_out["force_constants"] = np.asarray(h_lst)
            hdf5_out["vibrational_frequencies"] = np.asarray(nu_lst)
            hdf5_out["box_energy"] = np.asarray(e_lst)

    @property
    def _thermo(self):
        if self._td is None:
            self._td = Thermodynamics(
                self.strain_lst,
                self["output/vibrational_frequencies"],
                self["output/box_energy"],
                self.hessian.volume,
            )
        return self._td

    def get_helmholtz_free_energy(self, temperature, strain=None):
        return self._thermo.get_helmholtz_free_energy(
            temperature=temperature, strain=strain
        )

    get_helmholtz_free_energy.__doc__ = (
        thermo_doc.replace("INPUT_ONE", "temperature")
        .replace("INPUT_TWO", "strain")
        .replace("OUTPUT", "Helmholtz free energy")
    )
    get_helmholtz_free_energy.__doc__ = get_helmholtz_free_energy.__doc__.split("\n")
    get_helmholtz_free_energy.__doc__[0] = "Helmholtz free energy"
    get_helmholtz_free_energy.__doc__ = "\n".join(get_helmholtz_free_energy.__doc__)

    def get_gibbs_free_energy(self, temperature, pressure):
        return self._thermo.get_gibbs_free_energy(
            temperature=temperature, pressure=pressure
        )

    get_gibbs_free_energy.__doc__ = get_helmholtz_free_energy.__doc__.replace(
        "Helmholtz", "Gibbs"
    ).replace("strain", "pressure")

    def get_volume(self, temperature, pressure):
        return self.hessian.volume * (
            1 + self.get_strain(temperature=temperature, pressure=pressure)
        )

    get_volume.__doc__ = (
        thermo_doc.replace("INPUT_ONE", "temperature")
        .replace("INPUT_TWO", "pressure")
        .replace("OUTPUT", "volume")
    )

    def get_strain(self, temperature, pressure):
        return self._thermo.get_strain(temperature=temperature, pressure=pressure)

    get_strain.__doc__ = (
        thermo_doc.replace("INPUT_ONE", "temperature")
        .replace("INPUT_TWO", "pressure")
        .replace("OUTPUT", "strain")
    )

    def get_pressure(self, temperature, strain):
        return self._thermo.get_pressure(temperature=temperature, strain=strain)

    get_pressure.__doc__ = (
        thermo_doc.replace("INPUT_ONE", "temperature")
        .replace("INPUT_TWO", "strain")
        .replace("OUTPUT", "pressure")
    )

    def validate_ready_to_run(self):
        if self.ref_job._generic_input["calc_mode"] != "static":
            raise ValueError(
                "Phonopy reference jobs should be static calculations, but got {}".format(
                    self.ref_job._generic_input["calc_mode"]
                )
            )
        if self.input["num_points"] > 1 and self.input["num_points"] < 4:
            raise ValueError("num_points must be 1 or larger than 3")
        super().validate_ready_to_run()


class Thermodynamics:
    def __init__(self, strain, nu, E, volume):
        """

        Args:
            strain ((n_snapshots, n_atoms, 3)-np.ndarray): List of strain values
            nu ((n_snapshots, 3 * n_atoms)-np.ndarray): Vibrational frequencies
            E ((n_snapshots,)-ndarray): Force free energy values
            volume (float): Volume
        """
        self.strain = strain
        self.nu = nu
        self.E = E
        self.volume = volume
        self._fit_mat = None
        self._temperature = None
        self._free_energy = None

    @property
    def coeff(self):
        if len(self.free_energy) < 4:
            raise ValueError("Volume dependence makes sense only when num_points > 5")
        return np.einsum("ni,nj->ji", self.fit_mat, self.free_energy)

    @property
    def free_energy(self):
        if self._free_energy is None:
            self._free_energy = np.array(
                [
                    get_helmholtz_free_energy(nu, self.temperature, E)
                    for nu, E in zip(self.nu, self.E)
                ]
            )
        return self._free_energy

    @staticmethod
    def _harmonize(T, V):
        if T is None or V is None:
            return T, V
        V = np.atleast_1d(V)
        T = np.atleast_1d(T) * np.ones_like(V)
        V = V * np.ones_like(T)
        return T, V

    @property
    def temperature(self):
        return self._temperature

    @temperature.setter
    def temperature(self, T):
        self._temperature = np.asarray([T]).flatten()
        self._free_energy = None

    @property
    def fit_mat(self):
        if self._fit_mat is None:
            s_ni = self.strain[:, np.newaxis] ** np.arange(4)
            S_inv = np.linalg.inv(np.einsum("ni,nk->ik", *2 * [s_ni]))
            self._fit_mat = np.einsum("ik,nk->ni", S_inv, s_ni)
        return self._fit_mat

    def _get_min_strain(self, pressure):
        c = self.coeff.T
        return (-c[2] + np.sqrt(c[2] ** 2 - 3 * c[3] * (c[1] + pressure.flatten()))) / (
            3 * c[3]
        )

    def get_helmholtz_free_energy(self, temperature, strain=None):
        temperature, strain = self._harmonize(temperature, strain)
        self.temperature = temperature
        if strain is None or len(self.strain) == 1:
            return self.free_energy.squeeze()
        return (
            np.einsum(
                "ik,ik->i",
                self.coeff,
                strain.flatten()[..., np.newaxis] ** np.arange(4),
            )
            .reshape(np.shape(temperature))
            .squeeze()
        )

    def get_gibbs_free_energy(self, temperature, pressure):
        temperature, pressure = self._harmonize(temperature, pressure)
        pV = (
            (pressure * self.volume * 1e9 * unit.Pa * unit.angstrom**3).to("eV")
        ).magnitude
        self.temperature = temperature
        strain = self._get_min_strain(pV)
        return np.einsum(
            "ik,ik->i", self.coeff, strain.flatten()[:, np.newaxis] ** np.arange(4)
        ).reshape(np.shape(temperature))

    def get_strain(self, temperature, pressure):
        temperature, pressure = self._harmonize(temperature, pressure)
        pV = (
            (pressure * self.volume * 1e9 * unit.Pa * unit.angstrom**3).to("eV")
        ).magnitude
        self.temperature = temperature
        return self._get_min_strain(pV).reshape(np.shape(temperature))

    def get_pressure(self, temperature, strain):
        temperature, strain = self._harmonize(temperature, strain)
        self.temperature = temperature
        pressure = (
            -np.einsum(
                "ik,ik->i",
                (self.coeff * np.arange(4))[:, 1:],
                np.array([strain]).flatten()[:, np.newaxis] ** np.arange(3),
            ).reshape(np.shape(strain))
            / self.volume
        )
        return (pressure / 1.0e9 * unit.eV / unit.angstrom**3).to(unit.Pa).magnitude



import abc
from copy import deepcopy
import enum
from typing import Optional, Callable, List, Generator, Tuple

from pyiron_base.interfaces.object import HasStorage

from pyiron_contrib.tinybase.storage import Storable, pickle_dump, pickle_load
from pyiron_contrib.tinybase.container import (
    AbstractInput,
    AbstractOutput,
    StorageAttribute,
)


class ReturnStatus:
    """
    Status of the calculation.
    """

    class Code(enum.Enum):
        DONE = "done"
        ABORTED = "aborted"
        WARNING = "warning"
        NOT_CONVERGED = "not_converged"

    def __init__(self, code, msg=None):
        self.code = code if not isinstance(code, str) else ReturnStatus.Code(code)
        self.msg = msg

    @classmethod
    def done(cls, msg=None):
        return cls(code=cls.Code.DONE, msg=msg)

    @classmethod
    def aborted(cls, msg=None):
        return cls(code=cls.Code.ABORTED, msg=msg)

    @classmethod
    def warning(cls, msg=None):
        return cls(code=cls.Code.WARNING, msg=msg)

    @classmethod
    def not_converged(cls, msg=None):
        return cls(code=cls.Code.NOT_CONVERGED, msg=msg)

    def __repr__(self):
        return f"ReturnStatus({self.code}, {self.msg})"

    def __str__(self):
        return f"{self.code}({self.msg})"

    def is_done(self) -> True:
        """
        Returns True if calculation was successful.
        """
        return self.code == self.Code.DONE


class AbstractTask(Storable, abc.ABC):
    """
    Basic unit of calculations.

    Subclasses must implement :meth:`._get_input()`, :meth:`._get_output()` and :meth:`._execute()` and generally supply
    their own :class:`.AbstractInput` and :class:`.AbstractOutput`.
    """

    def __init__(self, capture_exceptions=True):
        self._input = None
        self._capture_exceptions = capture_exceptions

    @abc.abstractmethod
    def _get_input(self) -> AbstractInput:
        """
        Return an instance of the input class.

        This is called once per life time of the task object on first access to :attr:`.input` and then saved.
        """
        pass

    @property
    def input(self) -> AbstractInput:
        if self._input is None:
            self._input = self._get_input()
        return self._input

    @abc.abstractmethod
    def _get_output(self) -> AbstractOutput:
        """
        Return an instance of the output class.

        This is called every time :meth:`.execute()` is called.
        """
        pass

    @abc.abstractmethod
    def _execute(self, output) -> Optional[ReturnStatus]:
        """
        Run the calculation.

        Every time this method is called a new instance of the output is created and passed as the argument.  This
        method should populate it.

        If no value is returned from the method, a return status of DONE is assumed implicitly.

        Args:
            output (:class:`.AbstractOutput`): instance returned by :meth:`._get_output()`.

        Returns:
            :class:`.ReturnStatus`: optional
        """
        pass

    def execute(self) -> Tuple[ReturnStatus, AbstractOutput]:
        if not self.input.check_ready():
            return ReturnStatus.aborted("Input not ready!")
        output = self._get_output()
        try:
            ret = self._execute(output)
            if ret is None:
                ret = ReturnStatus("done")
        except Exception as e:
            ret = ReturnStatus("aborted", msg=e)
            if not self._capture_exceptions:
                raise
        return ret, output

    # TaskIterator Impl'
    def __iter__(
        self,
    ) -> Generator[
        List["Task"],
        List[Tuple[ReturnStatus, AbstractOutput]],
        Tuple[ReturnStatus, AbstractOutput],
    ]:
        ret, *_ = yield [self]
        return ret

    # Storable Impl'
    # We might even avoid this by deriving from HasStorage and put _input in there
    def _store(self, storage):
        # right now not all ASE objects can be stored in HDF, so let's just pickle for now
        storage["input"] = pickle_dump(self.input)
        # self.input.store(storage, "input")

    @classmethod
    def _restore(cls, storage, version):
        task = cls()
        task._input = pickle_load(storage["input"])
        return task


class TaskGenerator(AbstractTask, abc.ABC):
    """
    A generator that yields collections of tasks that can be executed in
    parallel.

    In each iteration it will yield a list of tasks and accept a list of their
    return status and outputs for the next iteration.  When it is done, it will
    return a final return status and output.
    """

    @abc.abstractmethod
    def __iter__(
        self,
    ) -> Generator[
        List["Task"],
        List[Tuple[ReturnStatus, AbstractOutput]],
        Tuple[ReturnStatus, AbstractOutput],
    ]:
        pass

    def _execute(self, output):
        gen = iter(self)
        tasks = next(gen)
        while True:
            ret = [t.execute() for t in tasks]
            try:
                tasks = gen.send(ret)
            except StopIteration as stop:
                ret, out = stop.args[0]
                output.take(out)
                return ret


# TaskGenerator.register(AbstractTask)
# assert


class FunctionInput(AbstractInput):
    args = StorageAttribute().type(list).constructor(list)
    kwargs = StorageAttribute().type(dict).constructor(dict)


class FunctionOutput(AbstractOutput):
    result = StorageAttribute()


class FunctionTask(AbstractTask):
    """
    A task that wraps a generic function.

    The given function is called with :attr:`.FunctionInput.args` and :attr:`.FunctionInput.kwargs` as `*args` and
    `**kwargs` respectively.  The return value is set to :attr:`.FunctionOutput.result`.
    """

    def __init__(self, function):
        super().__init__()
        self._function = function

    def _get_input(self):
        return FunctionInput()

    def _get_output(self):
        return FunctionOutput()

    def _execute(self, output):
        output.result = self._function(*self.input.args, **self.input.kwargs)


class ListInput(abc.ABC):
    """
    The input of :class:`.ListTaskGenerator`.

    To use it overload :meth:`._create_tasks()` here and subclass :class:`.ListTaskGenerator` as well.
    """

    @abc.abstractmethod
    def _create_tasks(self):
        """
        Return a list of tasks to execute.

        This is called once by :class:`.ListNode.execute`.
        """
        pass


class ListTaskGenerator(TaskGenerator, abc.ABC):
    """
    A task that executes other tasks in parallel.

    To use it overload :meth:`._extract_output` here and subclass :class:`.ListInput` as well.
    """

    def _get_input(self):
        return ListInput()

    @abc.abstractmethod
    def _extract_output(self, output, step, task, ret, task_output):
        """
        Extract the output of each task.

        Args:
            output (:class:`.AbstractOutput`): output of this task to populate
            step (int): index of the task to extract the output from,
            corresponds to the index of the task in the list returned by :meth:`.ListInput._create_tasks()`.
            task (:class:`.AbstractTask`): task to extract the output from, you can use this to extract parts of the input as well
            ret (:class:`.ReturnStatus`): the return status of the execution of the task
            task_output (:class:`.AbstractOutput`): the output of the task to extract
        """
        pass

    def __iter__(self):
        tasks = self.input._create_tasks()
        returns, outputs = zip(*(yield tasks))

        output = self._get_output()
        for i, (task, ret, task_output) in enumerate(zip(tasks, returns, outputs)):
            self._extract_output(output, i, task, ret, task_output)

        return ReturnStatus("done"), output


class SeriesInput(AbstractInput):
    """
    Keeps a list of tasks and their connection functions to run sequentially.

    The number of added tasks must be equal to the number of connections plus one.  It's recommended to set up this
    input with :meth:`.first()` and :meth:`.then()` which can be composed in a builder pattern.  The connection
    functions take as arguments the output of the last task and the input of the next task.  You may call
    :meth:`.then()` any number of times.

    >>> task = SeriesNode()
    >>> def transfer(input, output):
    ...     input.my_param = output.my_result
    >>> task.input.first(MyNode()).then(MyNode(), transfer)
    """

    tasks = StorageAttribute().type(list)
    connections = StorageAttribute().type(list)

    def check_ready(self):
        return len(self.tasks) == len(self.connections) + 1

    def first(self, task):
        """
        Set initial task.

        Resets whole input

        Args:
            task (AbstractTask): the first task to execute

        Returns:
            self: the input object
        """
        self.tasks = [task]
        self.connections = []
        return self

    def then(self, next_task, connection):
        """
        Add a new task and how to connect it to the previous task.

        Args:
            next_task (:class:`~.AbstractTask`): next task to execute
            connection (function): takes the input of next_task and the output
            of the previous task
        Returns:
            self: the input object
        """
        self.tasks.append(next_task)
        self.connections.append(connection)
        return self


class SeriesTask(TaskGenerator):
    """
    Executes a series of tasks sequentially.

    Its input specifies the tasks to execute and functions (:attr:`.SeriesInput.connections`) to move input from one
    output to the next input in the series.
    """

    def _get_input(self):
        return SeriesInput()

    def _get_output(self):
        return self.input.tasks[-1]._get_output()

    def __iter__(self):
        (ret, out), *_ = yield [self.input.tasks[0]]
        if not ret.is_done():
            return ReturnStatus("aborted", ret), None

        for task, connection in zip(self.input.tasks[1:], self.input.connections):
            connection(task.input, out)
            (ret, out), *_ = yield [task]
            if not ret.is_done():
                return ReturnStatus("aborted", ret), None

        return ret, out


class LoopControl(HasStorage):
    def __init__(self, condition, restart):
        super().__init__()
        self._condition = condition
        self._restart = restart

    scratch = StorageAttribute().constructor(dict)

    def condition(self, task: AbstractTask, output: AbstractTask):
        """
        Whether to terminate the loop or not.

        Args:
            task (AbstractTask): the loop body
            output (AbstractOutput): output of the loop body

        Args:
            bool: True to terminate the loop; False to keep it running
        """
        return self._condition(task, output, self.scratch)

    def restart(self, output: AbstractOutput, input: AbstractInput):
        """
        Setup the input of the next loop body.

        Args:
            output (AbstractOutput): output of the previous loop body
            input (AbstractInput): input of the next loop body
        """
        self._restart(output, input, self.scratch)


class RepeatLoopControl(LoopControl):
    def __init__(self, steps, restart=lambda *_: None):
        super().__init__(condition=self._count_steps, restart=restart)
        self.storage.steps = steps
        self.storage.counter = 0

    def _count_steps(self, output, input, scratch):
        c = self.storage.counter
        self.storage.counter += 1
        return c >= self.storage.steps


class LoopInput(AbstractInput):
    """
    Input for :class:`~.LoopTask`.

    Attributes:
        task (:class:`~.AbstractTask`): the loop body
        control (:class:`.LoopControl`): encapsulates control flow of the loop
    """

    trace = StorageAttribute().type(bool).default(False)
    control = StorageAttribute().type(LoopControl)

    def repeat(
        self,
        steps: int,
        restart: Optional[Callable[[AbstractOutput, AbstractInput, dict], None]] = None,
    ):
        """
        Set up a loop control that loops for steps and calls restart in between.

        If restart is not given don't do anything to input of the loop body.
        """
        if restart is not None:
            self.control = RepeatLoopControl(steps, restart)
        else:
            self.control = RepeatLoopControl(steps)

    def control_with(
        self,
        condition: Callable[[AbstractTask, AbstractOutput, dict], bool],
        restart: Callable[[AbstractOutput, AbstractInput, dict], None],
    ):
        """
        Set up a loop control that uses the callables for control flow.

        Args:
            condition (function): takes the loop body, its output and a persistant dict
            restart (function): takes the output of the last loop body, the input of the next one and a persistant dict
        """
        self.control = LoopControl(condition, restart)


class LoopTask(TaskGenerator):
    """
    Generic task to loop over a given input task.
    """

    def _get_input(self):
        return LoopInput()

    def _get_output(self):
        return self.input.task._get_output()

    def __iter__(self):
        task = deepcopy(self.input.task)
        control = deepcopy(self.input.control)

        while True:
            (ret, out), *_ = yield [task]
            if not ret.is_done():
                return ReturnStatus("aborted", ret), None
            if control.condition(task, out):
                break
            control.restart(out, task.input)
        return ret, out



"""
A base class for objects that can form nodes in the graph representation of a
computational workflow.
"""

from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from concurrent.futures import Future
from typing import Any, Optional, TYPE_CHECKING

from pyiron_contrib.executors import CloudpickleProcessPoolExecutor
from pyiron_contrib.workflow.files import DirectoryObject
from pyiron_contrib.workflow.has_to_dict import HasToDict
from pyiron_contrib.workflow.io import Signals, InputSignal, OutputSignal

if TYPE_CHECKING:
    from pyiron_base.jobs.job.extension.server.generic import Server

    from pyiron_contrib.workflow.composite import Composite
    from pyiron_contrib.workflow.io import Inputs, Outputs


class Node(HasToDict, ABC):
    """
    Nodes are elements of a computational graph.
    They have input and output data channels that interface with the outside
    world, and a callable that determines what they actually compute, and input and
    output signal channels that can be used to customize the execution flow of the
    graph;
    Together these channels represent edges on the computational graph.

    Nodes can be run to force their computation, or more gently updated, which will
    trigger a run only if the `run_on_update` flag is set to true and all of the input
    is ready (i.e. channel values conform to any type hints provided).

    Nodes may have a `parent` node that owns them as part of a sub-graph.

    Every node must be named with a `label`, and may use this label to attempt to create
    a working directory in memory for itself if requested.
    These labels also help to identify nodes in the wider context of (potentially
    nested) computational graphs.

    By default, nodes' signals input comes with `run` and `ran` IO ports which force
    the `run()` method and which emit after `finish_run()` is completed, respectfully.

    The `run()` method returns a representation of the node output (possible a futures
    object, if the node is running on an executor), and consequently `update()` also
    returns this output if the node is `ready` and has `run_on_updates = True`.

    Calling an already instantiated node allows its input channels to be updated using
    keyword arguments corresponding to the channel labels, performing a batch-update of
    all supplied input and then calling `update()`.
    As such, calling the node _also_ returns a representation of the output (or `None`
    if the node is not set to run on updates, or is otherwise unready to run).

    Nodes have a status, which is currently represented by the `running` and `failed`
    boolean flags.
    Their value is controlled automatically in the defined `run` and `finish_run`
    methods.

    Nodes can be run on the main python process that owns them, or by assigning an
    appropriate executor to their `executor` attribute.
    In case they are run with an executor, their `future` attribute will be populated
    with the resulting future object.

    This is an abstract class.
    Children *must* define how `inputs` and `outputs` are constructed, and what will
    happen `on_run`.
    They may also override the `run_args` property to specify input passed to the
    defined `on_run` method, and may add additional signal channels to the signals IO.

    # TODO: Everything with (de)serialization and executors for running on something
    #       other than the main python process.

    Attributes:
        connected (bool): Whether _any_ of the IO (including signals) are connected.
        failed (bool): Whether the node raised an error calling `run`. (Default
            is False.)
        fully_connected (bool): whether _all_ of the IO (including signals) are
            connected.
        future (concurrent.futures.Future | None): A futures object, if the node is
            currently running or has already run using an executor.
        inputs (pyiron_contrib.workflow.io.Inputs): **Abstract.** Children must define
            a property returning an `Inputs` object.
        label (str): A name for the node.
        outputs (pyiron_contrib.workflow.io.Outputs): **Abstract.** Children must define
            a property returning an `Outputs` object.
        parent (pyiron_contrib.workflow.composite.Composite | None): The parent object
            owning this, if any.
        ready (bool): Whether the inputs are all ready and the node is neither
            already running nor already failed.
        run_on_updates (bool): Whether to run when you are updated and all your input
            is ready and your status does not prohibit running. (Default is False).
        running (bool): Whether the node has called `run` and has not yet
            received output from this call. (Default is False.)
        server (Optional[pyiron_base.jobs.job.extension.server.generic.Server]): A
            server object for computing things somewhere else. Default (and currently
            _only_) behaviour is to compute things on the main python process owning
            the node.
        signals (pyiron_contrib.workflow.io.Signals): A container for input and output
            signals, which are channels for controlling execution flow. By default, has
            a `signals.inputs.run` channel which has a callback to the `run` method,
            and `signals.outputs.ran` which should be called at when the `run` method
            is finished.
            Additional signal channels in derived classes can be added to
            `signals.inputs` and  `signals.outputs` after this mixin class is
            initialized.

    Methods:
        disconnect: Remove all connections, including signals.
        on_run: **Abstract.** Do the thing.
        run: A wrapper to handle all the infrastructure around executing `on_run`.
    """

    def __init__(
        self,
        label: str,
        *args,
        parent: Optional[Composite] = None,
        run_on_updates: bool = False,
        **kwargs,
    ):
        """
        A mixin class for objects that can form nodes in the graph representation of a
        computational workflow.

        Args:
            label (str): A name for this node.
            *args: Arguments passed on with `super`.
            **kwargs: Keyword arguments passed on with `super`.

        TODO: Shouldn't `update_on_instantiation` and `run_on_updates` both live here??
        """
        super().__init__(*args, **kwargs)
        self.label: str = label
        self.parent = parent
        if parent is not None:
            parent.add(self)
        self.running = False
        self.failed = False
        # TODO: Replace running and failed with a state object
        self._server: Server | None = (
            None  # Or "task_manager" or "executor" -- we'll see what's best
        )
        # TODO: Move from a traditional "sever" to a tinybase "executor"
        # TODO: Provide support for actually computing stuff with the server/executor
        self.signals = self._build_signal_channels()
        self._working_directory = None
        self.run_on_updates: bool = run_on_updates
        self.executor: None | CloudpickleProcessPoolExecutor = None
        self.future: None | Future = None

    @property
    @abstractmethod
    def inputs(self) -> Inputs:
        pass

    @property
    @abstractmethod
    def outputs(self) -> Outputs:
        pass

    @property
    @abstractmethod
    def on_run(self) -> callable[..., Any | tuple]:
        """
        What the node actually does!
        """
        pass

    @property
    def run_args(self) -> dict:
        """
        Any data needed for `on_run`, will be passed as **kwargs.
        """
        return {}

    def process_run_result(self, run_output: Any | tuple) -> None:
        """
        What to _do_ with the results of `on_run` once you have them.

        Args:
            run_output (tuple): The results of a `self.on_run(self.run_args)` call.
        """
        pass

    def run(self) -> Any | tuple | Future:
        """
        Executes the functionality of the node defined in `on_run`.
        Handles the status of the node, and communicating with any remote
        computing resources.
        """
        if self.running:
            raise RuntimeError(f"{self.label} is already running")

        self.running = True
        self.failed = False

        if self.executor is None:
            try:
                run_output = self.on_run(**self.run_args)
            except Exception as e:
                self.running = False
                self.failed = True
                raise e
            return self.finish_run(run_output)
        elif isinstance(self.executor, CloudpickleProcessPoolExecutor):
            self.future = self.executor.submit(self.on_run, **self.run_args)
            self.future.add_done_callback(self.finish_run)
            return self.future
        else:
            raise NotImplementedError(
                "We currently only support executing the node functionality right on "
                "the main python process or with a "
                "pyiron_contrib.workflow.util.CloudpickleProcessPoolExecutor."
            )

    def finish_run(self, run_output: tuple | Future) -> Any | tuple:
        """
        Switch the node status, process the run result, then fire the ran signal.

        By extracting this as a separate method, we allow the node to pass the actual
        execution off to another entity and release the python process to do other
        things. In such a case, this function should be registered as a callback
        so that the node can finish "running" and, e.g. push its data forward when that
        execution is finished. In such a case, a `concurrent.futures.Future` object is
        expected back and must be unpacked.
        """
        if isinstance(run_output, Future):
            run_output = run_output.result()

        self.running = False
        try:
            self.process_run_result(run_output)
            self.signals.output.ran()
            return run_output
        except Exception as e:
            self.failed = True
            raise e

    def _build_signal_channels(self) -> Signals:
        signals = Signals()
        signals.input.run = InputSignal("run", self, self.run)
        signals.output.ran = OutputSignal("ran", self)
        return signals

    def update(self) -> Any | tuple | Future | None:
        if self.run_on_updates and self.ready:
            return self.run()

    @property
    def working_directory(self):
        if self._working_directory is None:
            if self.parent is not None and hasattr(self.parent, "working_directory"):
                parent_dir = self.parent.working_directory
                self._working_directory = parent_dir.create_subdirectory(self.label)
            else:
                self._working_directory = DirectoryObject(self.label)
        return self._working_directory

    @property
    def server(self) -> Server | None:
        return self._server

    @server.setter
    def server(self, server: Server | None):
        self._server = server

    def disconnect(self):
        self.inputs.disconnect()
        self.outputs.disconnect()
        self.signals.disconnect()

    @property
    def ready(self) -> bool:
        return not (self.running or self.failed) and self.inputs.ready

    @property
    def connected(self) -> bool:
        return self.inputs.connected or self.outputs.connected or self.signals.connected

    @property
    def fully_connected(self):
        return (
            self.inputs.fully_connected
            and self.outputs.fully_connected
            and self.signals.fully_connected
        )

    def _batch_update_input(self, **kwargs):
        """
        Temporarily disable running on updates to set all input values at once.

        Args:
            **kwargs: input label - input value (including channels for connection)
             pairs.
        """
        run_on_updates, self.run_on_updates = self.run_on_updates, False
        for k, v in kwargs.items():
            if k in self.inputs.labels:
                self.inputs[k] = v
            else:
                warnings.warn(
                    f"The keyword '{k}' was not found among input labels. If you are "
                    f"trying to update a node keyword, please use attribute assignment "
                    f"directly instead of calling, e.g. "
                    f"`my_node_instance.run_on_updates = False`."
                )
        self.run_on_updates = run_on_updates  # Restore provided value

    def __call__(self, **kwargs) -> None:
        self._batch_update_input(**kwargs)
        return self.update()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

__author__ = "Sam Waseda"
__copyright__ = (
    "Copyright 2023, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sam Waseda"
__email__ = "waseda@mpie.de"
__status__ = "production"
__date__ = "April 18, 2023"

import pandas as pd
from pyiron_atomistics.lammps.potential import LammpsPotentialFile
import numpy as np
import warnings
import itertools
from functools import wraps


general_doc = """
How to combine potentials:

Example I: Hybrid potential for a single element

>>> from pyiron_atomistics.lammps.potentials import Library, Morse
>>> eam = Library("Al")
>>> morse = Morse("Al", D_0=0.5, alpha=1.1, r_0=2.1, cutoff=6)
>>> lammps_job.potential = eam + morse

Example II: Hybrid potential for multiple elements

>>> from pyiron_atomistics.lammps.potentials import Library, Morse
>>> eam = Library("Al")
>>> morse_Al_Ni = Morse("Al", "Ni", D_0=0.2, alpha=1.05, r_0=2.2, cutoff=6)
>>> morse_Ni = Morse("Ni", D_0=0.7, alpha=1.15, r_0=2.15, cutoff=6)
>>> lammps_job.potential = eam + morse_Al_Ni + morse_Ni  # hybrid/overlay
>>> lammps_job.potential = eam * morse_Al_Ni * morse_Ni  # hybrid
>>> lammps_job.potential = 0.4 * eam + 0.1 * morse_Al_Ni + morse_Ni  # hybrid/scaled

"""


doc_pyiron_df = """
- "Config": Lammps commands in a list. Lines are separated by list items and
    each entry must finish with a new line
- "Filename": Potential file name in either absolute path or relative to
    the pyiron-resources path (optional)
- "Model": Model name (optional)
- "Name": Name of the potential (optional)
- "Species": Order of species as defined in pair_coeff
- "Citations": Citations (optional)

Example

>>> import pandas as pd
>>> pd.DataFrame(
...     {
...         "Config": [[
...             'pair_style my_potential 3.2\n',
...             'pair_coeff 2 1 1.1 2.3 3.2\n'
...         ]],
...         "Filename": [""],
...         "Model": ["my_model"],
...         "Name": ["my_potential"],
...         "Species": [["Fe", "Al"]],
...         "Citations": [],
...     }
... )
"""


issue_page = "https://github.com/pyiron/pyiron_atomistics/issues"


class LammpsPotentials:
    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        cls._df = None
        return obj

    @staticmethod
    def _harmonize_species(species_symbols) -> list:
        """
        Check whether species are set for the pairwise interactions. If only
        one chemical species is given, duplicate the species.
        """
        if len(species_symbols) == 0:
            raise ValueError("Chemical elements not specified")
        if len(species_symbols) == 1:
            species_symbols *= 2
        return list(species_symbols)

    def _initialize_df(
        self,
        pair_style,
        interacting_species,
        pair_coeff,
        preset_species=None,
        model=None,
        citations=None,
        filename=None,
        potential_name=None,
        scale=None,
        cutoff=None,
    ):
        def check_none_n_length(variable, default, length=len(pair_coeff)):
            if variable is None:
                variable = default
            if isinstance(variable, list) and len(variable) == 1 < length:
                variable = length * variable
            return variable

        arg_dict = {
            "pair_style": pair_style,
            "interacting_species": interacting_species,
            "pair_coeff": pair_coeff,
            "preset_species": check_none_n_length(preset_species, [[]]),
            "cutoff": check_none_n_length(cutoff, 0),
            "model": check_none_n_length(model, pair_style),
            "citations": check_none_n_length(citations, [[]]),
            "filename": check_none_n_length(filename, [""]),
            "potential_name": check_none_n_length(potential_name, pair_style),
        }
        if scale is not None:
            arg_dict["scale"] = scale
        try:
            self.set_data(pd.DataFrame(arg_dict))
        except ValueError:
            raise ValueError(
                f"Initialization failed - inconsistency in data: {arg_dict}"
            )

    def copy(self):
        new_pot = LammpsPotentials()
        new_pot.set_data(self.get_all_data())
        return new_pot

    @staticmethod
    def _unique(args):
        labels, indices = np.unique(args, return_index=True)
        return labels[np.argsort(indices)]

    @property
    def model(self) -> str:
        """Model name (required in pyiron df)"""
        return "_and_".join(self._unique(self.df.model))

    @property
    def potential_name(self) -> str:
        """Potential name (required in pyiron df)"""
        return "_and_".join(self._unique(self.df.potential_name))

    @property
    def species(self):
        """Species defined in the potential"""
        species = self._unique([ss for s in self.df.interacting_species for ss in s])
        preset = self._unique(
            ["___".join(s) for s in self.df.preset_species if len(s) > 0]
        )
        if len(preset) == 0:
            return list(species)
        elif len(preset) > 1:
            raise NotImplementedError(
                "Currently not possible to have multiple file-based potentials"
            )
        preset = list(preset)[0].split("___")
        comp_lst = [s for s in species if s not in self._unique(preset)]
        return [p for p in preset + comp_lst if p != "*"]

    @property
    def filename(self) -> list:
        """LAMMPS potential files"""
        return [f for f in self._unique(self.df.filename) if len(f) > 0]

    @property
    def citations(self) -> str:
        """Citations to be included"""
        return "".join(np.unique([c for c in self.df.citations if len(c) > 0]))

    @property
    def is_scaled(self) -> bool:
        """Scaling in pair_style hybrid/scaled and hybrid/overlay (which is scale=1)"""
        return "scale" in self.df

    @property
    def pair_style(self) -> str:
        """LAMMPS pair_style"""
        if len(set(self.df.pair_style)) == 1:
            pair_style = "pair_style " + list(set(self.df.pair_style))[0]
            if np.max(self.df.cutoff) > 0:
                pair_style += f" {np.max(self.df.cutoff)}"
            return pair_style + "\n"
        elif "scale" not in self.df:
            pair_style = "pair_style hybrid"
        elif all(self.df.scale == 1):
            pair_style = "pair_style hybrid/overlay"
        else:
            pair_style = "pair_style hybrid/scaled"
        for ii, s in enumerate(self.df[["pair_style", "cutoff"]].values):
            if pair_style.startswith("pair_style hybrid/scaled"):
                pair_style += f" {self.df.iloc[ii].scale}"
            pair_style += f" {s[0]}"
            if s[1] > 0:
                pair_style += f" {s[1]}"
        return pair_style + "\n"

    class _PairCoeff:
        def __init__(
            self,
            pair_style,
            interacting_species,
            pair_coeff,
            species,
            preset_species,
        ):
            self._interacting_species = interacting_species
            self._pair_coeff = pair_coeff
            self._species = species
            self._preset_species = preset_species
            self._pair_style = pair_style
            self._s_dict = None

        @property
        def is_hybrid(self):
            return len(set(self._pair_style)) > 1

        @property
        def counter(self):
            """
            Enumeration of potentials if a potential is used multiple
            times in hybrid (which is a requirement from LAMMPS)
            """
            key, count = np.unique(self._pair_style, return_counts=True)
            counter = {kk: 1 for kk in key[count > 1]}
            results = []
            for coeff in self._pair_style:
                if coeff in counter and self.is_hybrid:
                    results.append(str(counter[coeff]))
                    counter[coeff] += 1
                else:
                    results.append("")
            return results

        @property
        def pair_style(self):
            """pair_style to be output only in hybrid"""
            if self.is_hybrid:
                return self._pair_style
            else:
                return len(self._pair_style) * [""]

        @property
        def results(self):
            """pair_coeff lines to be used in pyiron df"""
            return [
                " ".join((" ".join(("pair_coeff",) + c)).split()) + "\n"
                for c in zip(
                    self.interacting_species,
                    self.pair_style,
                    self.counter,
                    self.pair_coeff,
                )
            ]

        @property
        def s_dict(self):
            if self._s_dict is None:
                self._s_dict = dict(
                    zip(self._species, (np.arange(len(self._species)) + 1).astype(str))
                )
                self._s_dict.update({"*": "*"})
            return self._s_dict

        @property
        def interacting_species(self) -> list:
            """
            Species in LAMMPS notation (i.e. in numbers instead of chemical
            symbols)
            """
            return [
                " ".join([self.s_dict[cc] for cc in c])
                for c in self._interacting_species
            ]

        @property
        def pair_coeff(self) -> list:
            """
            Args for pair_coeff. Elements defined in EAM files are
            complemented with the ones defined in other potentials in the
            case of hybrid (filled with NULL)
            """
            if not self.is_hybrid:
                return self._pair_coeff
            results = []
            for pc, ps in zip(self._pair_coeff, self._preset_species):
                if len(ps) > 0 and "eam" in pc:
                    s = " ".join(ps + (len(self._species) - len(ps)) * ["NULL"])
                    pc = s.join(pc.rsplit(" ".join(ps), 1))
                results.append(pc)
            return results

    @property
    def pair_coeff(self) -> list:
        """LAMMPS pair_coeff"""

        return self._PairCoeff(
            pair_style=self.df.pair_style,
            interacting_species=self.df.interacting_species,
            pair_coeff=self.df.pair_coeff,
            species=self.species,
            preset_species=self.df.preset_species,
        ).results

    def get_df(self):
        """df used in pyiron potential"""
        return pd.DataFrame(
            {
                "Config": [[self.pair_style] + self.pair_coeff],
                "Filename": [self.filename],
                "Model": [self.model],
                "Name": [self.potential_name],
                "Species": [self.species],
                "Citations": [self.citations],
            }
        )

    def __repr__(self):
        return self.df.__repr__()

    def _repr_html_(self):
        return self.df._repr_html_()

    def set_data(self, df):
        for key in [
            "pair_style",
            "interacting_species",
            "pair_coeff",
            "preset_species",
        ]:
            if key not in df:
                raise ValueError(f"{key} missing")
        self._df = df

    @property
    def df(self):
        """DataFrame containing all info for each pairwise interactions"""
        return self._df

    def get_all_data(self, default_scale=None):
        if default_scale is None or "scale" in self.df:
            return self.df.copy()
        df = self.df.copy()
        df["scale"] = default_scale
        return df

    def __mul__(self, scale_or_potential):
        if isinstance(scale_or_potential, LammpsPotentials):
            if self.is_scaled or scale_or_potential.is_scaled:
                raise ValueError("You cannot mix hybrid types")
            new_pot = LammpsPotentials()
            new_pot.set_data(
                pd.concat(
                    (self.get_all_data(), scale_or_potential.get_all_data()),
                    ignore_index=True,
                )
            )
            return new_pot
        if self.is_scaled:
            raise NotImplementedError("Currently you cannot scale twice")
        new_pot = self.copy()
        new_pot.df["scale"] = scale_or_potential
        return new_pot

    __rmul__ = __mul__

    def __add__(self, potential):
        new_pot = LammpsPotentials()
        new_pot.set_data(
            pd.concat(
                (
                    self.get_all_data(default_scale=1),
                    potential.get_all_data(default_scale=1),
                ),
                ignore_index=True,
            )
        )
        return new_pot


class Library(LammpsPotentials):
    """
    Potential class to choose a file based potential from an existing library
    (e.g. EAM).
    You can either specify the chemical species and/or the name of the
    potential.

    Example I: Via chemical species

    >>> eam = Library("Al")

    Example II: Via potential name

    >>> eam = Library(name="1995--Angelo-J-E--Ni-Al-H--LAMMPS--ipr1")

    If the variable `eam` is used without specifying the potential name (i.e.
    in Example I), the first potential in the database corresponding with the
    specified chemical species will be selected. In order to see the list of
    potentials, you can also execute

    >>> eam = Library("Al")
    >>> eam.list_potentials()  # See list of potential names
    >>> eam.view_potentials()  # See potential names and metadata

    """

    def __init__(self, *chemical_elements, name=None):
        """
        Args:
            chemical_elements (str): chemical elements/species
            name (str): potential name in the database
        """
        if name is not None:
            self._df_candidates = LammpsPotentialFile().find_by_name(name)
        else:
            self._df_candidates = LammpsPotentialFile().find(list(chemical_elements))

    @staticmethod
    def _get_pair_style(config):
        if any(["hybrid" in c for c in config]):
            return [c.split()[3] for c in config if "pair_coeff" in c]
        for c in config:
            if "pair_style" in c:
                return [" ".join(c.replace("\n", "").split()[1:])] * sum(
                    ["pair_coeff" in c for c in config]
                )
        raise ValueError(
            f"pair_style could not determined: {config}.\n\n"
            "The reason why you are seeing this error is most likely because "
            "the potential you chose had a corrupt config. It is "
            "supposed to have at least one item which starts with 'pair_style'.\n"
            "If you are using the standard pyiron database, feel free to "
            f"submit an issue on {issue_page}. "
            "Typically you can get a reply within 24h.\n"
        )

    @staticmethod
    def _get_pair_coeff(config):
        try:
            if any(["hybrid" in c for c in config]):
                return [" ".join(c.split()[4:]) for c in config if "pair_coeff" in c]
            return [" ".join(c.split()[3:]) for c in config if "pair_coeff" in c]
        except IndexError:
            raise AssertionError(
                f"{config} does not follow the format 'pair_coeff element_1 element_2 args'"
            )

    @staticmethod
    def _get_interacting_species(config, species):
        def _convert(c, s):
            if c == "*":
                return c
            return s[int(c) - 1]

        return [
            [_convert(cc, species) for cc in c.split()[1:3]]
            for c in config
            if c.startswith("pair_coeff")
        ]

    @staticmethod
    def _get_scale(config):
        for c in config:
            if not c.startswith("pair_style"):
                continue
            if "hybrid/overlay" in c:
                return 1
            elif "hybrid/scaled" in c:
                raise NotImplementedError(
                    "Too much work for something inexistent in pyiron database for now"
                )
        return

    def list_potentials(self):
        return self._df_candidates.Name

    def view_potentials(self):
        return self._df_candidates

    @property
    def df(self):
        if self._df is None:
            df = self._df_candidates.iloc[0]
            if len(self._df_candidates) > 1:
                warnings.warn(
                    f"Potential not uniquely specified - use default {df.Name}"
                )
            self._initialize_df(
                pair_style=self._get_pair_style(df.Config),
                interacting_species=self._get_interacting_species(
                    df.Config, df.Species
                ),
                pair_coeff=self._get_pair_coeff(df.Config),
                preset_species=[df.Species],
                model=df.Model,
                citations=df.Citations,
                filename=df.Filename,
                potential_name=df.Name,
                scale=self._get_scale(df.Config),
            )
        return self._df


def check_cutoff(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if "cutoff" not in kwargs or kwargs["cutoff"] == 0:
            raise ValueError(
                "It is not possible to set cutoff=0 for parameter-based "
                "potentials. If you think this should be possible, you have the "
                "following options:\n\n"
                f"- Open an issue on our GitHub page: {issue_page}\n"
                "- Write your own potential in pyiron format. Here's how:\n"
                f"{doc_pyiron_df}\n"
            )
        return f(*args, **kwargs)

    return wrapper


class Morse(LammpsPotentials):
    """
    Morse potential defined by:

    E = D_0*[exp(-2*alpha*(r-r_0))-2*exp(-alpha*(r-r_0))]
    """

    @check_cutoff
    def __init__(self, *chemical_elements, D_0, alpha, r_0, cutoff, pair_style="morse"):
        """
        Args:
            chemical_elements (str): Chemical elements
            D_0 (float): parameter (s. eq. above)
            alpha (float): parameter (s. eq. above)
            r_0 (float): parameter (s. eq. above)
            cutoff (float): cutoff length
            pair_style (str): pair_style name (default: "morse")

        Example:

        >>> morse = Morse("Al", "Ni", D_0=1, alpha=0.5, r_0=2, cutoff=6)
        """
        self._initialize_df(
            pair_style=[pair_style],
            interacting_species=[self._harmonize_species(chemical_elements)],
            pair_coeff=[" ".join([str(cc) for cc in [D_0, alpha, r_0, cutoff]])],
            cutoff=cutoff,
        )


Morse.__doc__ += general_doc


class CustomPotential(LammpsPotentials):
    """
    Custom potential class to define LAMMPS potential not implemented in
    pyiron
    """

    @check_cutoff
    def __init__(self, pair_style, *chemical_elements, cutoff, **kwargs):
        """
        Args:
            pair_style (str): pair_style name (default: "morse")
            chemical_elements (str): Chemical elements
            cutoff (float): cutoff length

        Example:

        >>> custom_pot = CustomPotential("lj/cut", "Al", "Ni", epsilon=0.5, sigma=1, cutoff=3)

        Important: the order of parameters is conserved in the LAMMPS input
        (except for `cutoff`, which is always the last argument).
        """
        self._initialize_df(
            pair_style=[pair_style],
            interacting_species=[self._harmonize_species(chemical_elements)],
            pair_coeff=[" ".join([str(cc) for cc in kwargs.values()]) + f" {cutoff}"],
            cutoff=cutoff,
        )


CustomPotential.__doc__ += general_doc



import abc
import os.path
from typing import Union

from pyiron_base import Project, DataContainer
from pyiron_contrib.tinybase.storage import (
    GenericStorage,
    ProjectHDFioStorageAdapter,
    DataContainerAdapter,
)
from pyiron_contrib.tinybase.database import TinyDB, GenericDatabase


class JobNotFoundError(Exception):
    pass


class ProjectInterface(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def open_location(cls, location) -> "ProjectInterface":
        pass

    @abc.abstractmethod
    def create_storage(self, name) -> GenericStorage:
        pass

    @abc.abstractmethod
    def exists_storage(self, name) -> bool:
        pass

    @abc.abstractmethod
    def remove_storage(self, name):
        pass

    @abc.abstractmethod
    def _get_database(self) -> GenericDatabase:
        pass

    @property
    def database(self):
        return self._get_database()

    @property
    def create(self):
        if not hasattr(self, "_creator"):
            from pyiron_contrib.tinybase.creator import RootCreator, CREATOR_CONFIG

            self._creator = RootCreator(self, CREATOR_CONFIG)
        return self._creator

    def load(self, name_or_id: Union[int, str]) -> "TinyJob":
        """
        Load a job from storage.

        If the job name is given, it must be a child of this project and not
        any of its sub projects.

        Args:
            name_or_id (int, str): either the job name or its id

        Returns:
            :class:`.TinyJob`: the loaded job

        Raises:
            :class:`.JobNotFoundError`: if no job of the given name or id exists
        """
        if isinstance(name_or_id, str):
            pr = self
            name = name_or_id
            if not pr.exists_storage(name):
                raise JobNotFoundError(f"No job with name {name} found!")
        else:
            try:
                entry = self.database.get_item(name_or_id)
            except ValueError as e:
                raise JobNotFoundError(*e.args)
            pr = self.open_location(entry.project)
            name = entry.name
        return pr.create_storage(name).to_object()

    @property
    @abc.abstractmethod
    def path(self):
        pass

    @property
    @abc.abstractmethod
    def name(self):
        pass

    def job_table(self):
        return self.database.job_table()

    def get_job_id(self, name):
        project_id = self.database.get_project_id(self.path)
        return self.database.get_item_id(name, project_id)

    def remove(self, job_id):
        entry = self.database.remove_item(job_id)
        if entry.project == self.path:
            pr = self
        else:
            pr = self.open_location(entry.project)
        pr.remove_storage(entry.name)

    # TODO:
    # def copy_to/move_to across types of ProjectInterface


class ProjectAdapter(ProjectInterface):
    def __init__(self, project):
        self._project = project
        self._database = None

    @classmethod
    def open_location(cls, location):
        return cls(Project(location))

    def create_storage(self, name):
        return ProjectHDFioStorageAdapter(
            self, self._project.create_hdf(self._project.path, name)
        )

    def exists_storage(self, name) -> bool:
        return self._project.create_hdf(self._project.path, name).file_exists

    def remove_storage(self, name):
        self._project.create_hdf(self._project.path, name).remove_file()

    def _get_database(self):
        if self._database is None:
            self._database = TinyDB(os.path.join(self._project.path, "pyiron.db"))
        return self._database

    @property
    def name(self):
        return self._project.name

    @property
    def path(self):
        return self._project.path


class InMemoryProject(ProjectInterface):
    def __init__(self, location, db=None, storage=None):
        if db is None:
            db = TinyDB(":memory:")
        self._db = db
        self._location = location
        if storage is None:
            storage = {}
        self._storage = storage
        if location not in storage:
            self._storage[location] = DataContainer()

    def open_location(self, location):
        return self.__class__(location, db=self.database, storage=self._storage)

    def create_storage(self, name) -> GenericStorage:
        return DataContainerAdapter(
            self, self._storage[self._location], "/"
        ).create_group(name)

    def exists_storage(self, name) -> bool:
        return name in self._storage[self._location].list_groups()

    def remove_storage(self, name):
        self._storage[self._location].pop(name)

    def _get_database(self) -> GenericDatabase:
        return self._db

    @property
    def path(self):
        return self._location

    @property
    def name(self):
        return os.path.basename(self.path)



"""
Inspects code to automatically parse return values as strings
"""

import ast
import inspect
import re
from textwrap import dedent


def _remove_spaces_until_character(string):
    pattern = r"\s+(?=\s)"
    modified_string = re.sub(pattern, "", string)
    return modified_string


class ParseOutput:
    """
    Given a function with at most one `return` expression, inspects the source code and
    parses a list of strings containing the returned values.
    If the function returns `None`, the parsed value is also `None`.
    This parsed value is evaluated at instantiation and stored in the `output`
    attribute.
    In case more than one `return` expression is found, a `ValueError` is raised.
    """

    def __init__(self, function):
        self._func = function
        self._source = None
        self._output = self.get_parsed_output()

    @property
    def func(self):
        return self._func

    @property
    def dedented_source_string(self):
        return dedent(inspect.getsource(self.func))

    @property
    def node_return(self):
        tree = ast.parse(self.dedented_source_string)
        returns = []
        for node in ast.walk(tree):
            if isinstance(node, ast.Return):
                returns.append(node)

        if len(returns) > 1:
            raise ValueError(
                f"{self.__class__.__name__} can only parse callables with at most one "
                f"return value, but ast.walk found {len(returns)}."
            )

        try:
            return returns[0]
        except IndexError:
            return None

    @property
    def source(self):
        if self._source is None:
            self._source = self.dedented_source_string.split("\n")[:-1]
        return self._source

    def get_string(self, node):
        string = ""
        for ll in range(node.lineno - 1, node.end_lineno):
            if ll == node.lineno - 1 == node.end_lineno - 1:
                string += _remove_spaces_until_character(
                    self.source[ll][node.col_offset : node.end_col_offset]
                )
            elif ll == node.lineno - 1:
                string += _remove_spaces_until_character(
                    self.source[ll][node.col_offset :]
                )
            elif ll == node.end_lineno - 1:
                string += _remove_spaces_until_character(
                    self.source[ll][: node.end_col_offset]
                )
            else:
                string += _remove_spaces_until_character(self.source[ll])
        return string

    @property
    def output(self):
        return self._output

    def get_parsed_output(self):
        if self.node_return is None or self.node_return.value is None:
            return
        elif isinstance(self.node_return.value, ast.Tuple):
            return [self.get_string(s) for s in self.node_return.value.dims]
        else:
            out = [self.get_string(self.node_return.value)]
            if out == ["None"]:
                return
            else:
                return out



import numpy as np


def setup_lmp_input(lmp, n_atoms=None, direction=None, fix_id=-1):
    """
    Change input for LAMMPS to run a drag calculation.

    Args:
        lmp (pyiron_atomistics.lammps.lammps.Lammps): LAMMPS job
        n_atoms (int): number of free atoms (default: None, i.e. it is
            determined from the job structure)
        direction (None/numpy.ndarray): direction along which the force is
            cancelled. None if all forces are to be cancelled (default: None)
        fix_id (None/int): id of the atom to be fixed (default: -1, i.e. last
            atom)

    Returns:
        None (input of lmp is changed in-place)

    In the context of this function, a drag calculation is a constraint energy
    minimization, in which one atom is either not allowed to move at all, or
    not allowed to move along a given direction. In order for the system to not
    fall to the energy minimum, the sum of the remaining forces is set to 0.

    Exp: Hydrogen diffusion

    >>> from pyiron_atomistics import Project
    >>> import numpy as np
    >>> pr = Project("DRAG")
    >>> bulk = pr.create.structure.bulk('Ni', cubic=True)
    >>> a_0 = bulk.cell[0, 0]
    >>> x_octa = np.array([0, 0, 0.5 * a_0])
    >>> x_tetra = np.array(3 * [0.25 * a_0])
    >>> dx = x_tetra - x_octa
    >>> transition = np.linspace(0, 1, 101)
    >>> x_lst = transition[:, None] * dx + x_octa
    >>> structure = bulk.repeat(4)
    >>> structure += pr.create.structure.atoms(
    ...     positions=[x_octa],
    ...     elements=['H'],
    ...     cell=structure.cell
    ... )
    >>> lmp = pr.create.job.Lammps('lmp')
    >>> lmp.structure = structure
    >>> lmp.calc_minimize()
    >>> lmp.potential = potential_of_your_choice
    >>> setup_lmp_input(lmp, direction=dx)
    >>> lmp.interactive_open()
    >>> for xx in x_lst:
    >>>     lmp.structure.positions[-1] = xx
    >>>     lmp.run()
    >>> lmp.interactive_close()
    """
    if lmp.input.control["minimize"] is None:
        raise ValueError("set calc_minimize first")
    if n_atoms is None:
        try:
            n_atoms = len(lmp.structure) - 1
        except TypeError:
            raise AssertionError("either `n_atoms` or the structure must be set")
    fix_id = np.arange(n_atoms + 1)[fix_id] + 1
    lmp.input.control["atom_modify"] = "map array"
    lmp.input.control["group___fixed"] = f"id {fix_id}"
    lmp.input.control["group___free"] = "subtract all fixed"
    if direction is None:
        for ii, xx in enumerate(["x", "y", "z"]):
            lmp.input.control[
                f"variable___f{xx}_free"
            ] = f"equal f{xx}[{fix_id}]/{n_atoms}"
            lmp.input.control[f"variable___f{xx}_fixed"] = f"equal -f{xx}[{fix_id}]"
    else:
        direction = np.array(direction) / np.linalg.norm(direction)
        direction = np.outer(direction, direction)
        direction = np.around(direction, decimals=8)
        for grp, ss in zip(["free", "fixed"], [f"1/{n_atoms}*", "-"]):
            for ii, xx in enumerate(["x", "y", "z"]):
                txt = "+".join(
                    [
                        f"({ss}f{xxx}[{fix_id}]*({direction[ii][iii]}))"
                        for iii, xxx in enumerate(["x", "y", "z"])
                        if not np.isclose(direction[ii][iii], 0)
                    ]
                )
                if txt == "":
                    txt = 0
                lmp.input.control[f"variable___f{xx}_{grp}"] = f" equal {txt}"
    lmp.input.control["variable___energy"] = "atom 0"
    for key in ["free", "fixed"]:
        txt = " ".join([f"v_f{x}_{key}" for x in ["x", "y", "z"]])
        lmp.input.control[f"fix___f_{key}"] = f"{key} addforce {txt} energy v_energy"
    lmp.input.control["min_style"] = "quickmin"



import os
import subprocess
from glob import glob

from pyiron_base.state import state

from pyiron_contrib.tinybase.container import (
    AbstractInput,
    AbstractOutput,
    StorageAttribute,
)

from pyiron_contrib.tinybase.task import AbstractTask, ReturnStatus

if os.name == "nt":
    EXE_SUFFIX = "bat"
else:
    EXE_SUFFIX = "sh"


class ExecutablePathResolver:
    """
    Locates executables in pyiron resource folders.

    This expects executables to be located in folder structures like this

    {resource_folder}/{module}/bin/run_{code}_{version}.sh

    and be executable (on UNIX).

    If multiple executables are found for the same combination of `module`,
    `code` and `version`, :meth:`.list()` returns all of them sorted first by
    resource path (as in PYIRON_RESOURCE_PATH) and then alphabetically.

    If an executable exists that has `default` in its version and `version` is
    `None`, the first such executable is picked for :meth:`.path()`.

    :meth:`.__str__` is overloaded to :meth:`.path()`.
    """

    def __init__(self, module, code, version=None):
        self._module = module
        self._code = code
        self._version = version

    def list(self, version=None):
        """
        List all possible executables found.

        Returns: list of str
        """
        if version is None:
            version = self._version or "*"
        alternatives = []
        for p in state.settings.resource_paths:
            exe_path = f"run_{self._code}_{version}.{EXE_SUFFIX}"
            bin_path = os.path.join(p, self._module, "bin", exe_path)
            alternatives.extend(sorted(glob(bin_path)))
        return alternatives

    def list_versions(self):
        """
        List unique version strings found.
        """
        exes = self.list(version="*")

        def extract(p):
            return os.path.splitext(
                os.path.basename(p).split(f"run_{self._code}", maxsplit=1)[1]
            )[0][1:]

        return list(set(map(extract, exes)))

    @property
    def version(self):
        vers = self.list_versions()
        for v in vers:
            if "default" in vers:
                return v
        return vers[0]

    @version.setter
    def version(self, value):
        vers = self.list_versions()
        if value in vers:
            self._version = value
        else:
            raise ValueError(f"Given version '{value}' not in {vers}!")

    def path(self):
        """
        Returns a direct path where a executable has been found.
        """
        exes = self.list()
        if self._version is not None:
            return exes[0]
        for p in exes:
            if "default" in p:
                return p
        return exes[0]

    def __str__(self):
        return self.path()


class ShellInput(AbstractInput):
    command = StorageAttribute()
    arguments = StorageAttribute().type(list).constructor(list)
    environ = StorageAttribute().type(dict).constructor(dict)
    working_directory = StorageAttribute().type(str)
    allowed_returncode = StorageAttribute().type(list)


class ShellOutput(AbstractOutput):
    stdout = StorageAttribute()
    stderr = StorageAttribute()
    returncode = StorageAttribute().type(int)


class ShellTask(AbstractTask):
    def _get_input(self):
        return ShellInput()

    def _get_output(self):
        return ShellOutput()

    def _execute(self, output):
        environ = dict(os.environ)
        environ.update({k: str(v) for k, v in self.input.environ.items()})
        proc = subprocess.run(
            [str(self.input.command), *map(str, self.input.arguments)],
            capture_output=True,
            cwd=self.input.working_directory,
            encoding="utf8",
            env=environ,
        )
        output.stdout = proc.stdout
        output.stderr = proc.stderr
        output.returncode = proc.returncode
        allowed_returncode = self.input.allowed_returncode
        if allowed_returncode is None:
            allowed_returncode = [0]
        if proc.returncode not in allowed_returncode:
            return ReturnStatus("aborted", f"non-zero error code {proc.returncode}")



import enum
from collections import defaultdict
from functools import partial
from typing import Union, List
import time
import logging
from math import inf

from pyiron_contrib.tinybase.task import AbstractTask, TaskGenerator


class RunMachine:
    class Code(enum.Enum):
        INIT = "init"
        READY = "ready"
        RUNNING = "running"
        COLLECT = "collect"
        FINISHED = "finished"

    def __init__(self, initial_state):
        self._state = RunMachine.Code(initial_state)
        self._callbacks = {}
        self._observers = defaultdict(list)
        self._data = {}  # state variables associated with each state

    @property
    def state(self):
        return self._state

    def on(self, state: Union[str, Code], callback):
        if isinstance(state, str):
            state = RunMachine.Code(state)
        self._callbacks[state] = callback

    def observe(self, state: Union[str, Code], callback):
        if isinstance(state, str):
            state = RunMachine.Code(state)
        self._observers[state].append(callback)

    def goto(self, state: Union[str, Code], **kwargs):
        if isinstance(state, str):
            state = RunMachine.Code(state)
        self._state = state
        self._data = {}
        self._data.update(kwargs)
        for obs in self._observers[state]:
            obs(self._data)

    def step(self, state: Union[str, Code, None] = None, **kwargs):
        if state is not None:
            self.goto(state, **kwargs)
        self._callbacks.get(self._state, lambda: None)()


class ExecutionContext:
    def __init__(self, tasks):
        self._tasks = tasks
        self._run_machine = RunMachine("init")

        # Set up basic flow
        self._run_machine.on("init", self._run_init)
        # exists mostly to let downstream code hook into it, e.g. to write input files and such
        self._run_machine.on("ready", self._run_ready)
        self._run_machine.on("running", self._run_running)
        # exists mostly to let downstream code hook into it, e.g. to read output files and such
        self._run_machine.on("collect", self._run_collect)
        self._run_machine.on("finished", self._run_finished)

        # keeping track of run times
        self._running_start = None
        self._collect_start = None
        self._finished_start = None
        self._run_machine.observe("running", self._time_running)
        self._run_machine.observe("collect", self._time_collect)
        self._run_machine.observe("finished", self._time_finished)

    def _time_running(self, _data):
        self._running_start = time.monotonic()

    def _time_collect(self, _data):
        self._collect_start = time.monotonic()
        self._run_time = self._collect_start - self._running_start

    def _time_finished(self, _data):
        self._finished_start = time.monotonic()
        self._collect_time = self._finished_start - self._collect_start

    @property
    def tasks(self):
        return self._tasks

    def _run_init(self):
        if all(task.input.check_ready() for task in self.tasks):
            self._run_machine.step("ready")
        else:
            logging.info("Task is not ready yet!")

    def _run_ready(self):
        self._run_machine.step("running")

    def _run_running(self):
        status, output = zip(*[task.execute() for task in self.tasks])
        self._run_machine.step("collect", status=status, output=output)

    def _run_collect(self):
        self._run_machine.step(
            "finished",
            status=self.status,
            output=self.output,
        )

    def _run_finished(self):
        pass

    def run(self):
        self._run_machine.step()

    def wait(self, until="finished", timeout=None, sleep=0.1):
        """
        Sleep until specified state of the run state machine is reached.

        Before calling this method, the state of this context must be past
        `init`, i.e. you have to call :meth:`.run` at least once.

        Args:
            until (str): wait until the executor has reached this state; must
                be a valid state name of :class:`.RunMachine.Code`.
            timeout (float): maximum amount of seconds to wait; wait
                indefinitely by default
            sleep (float): amount of seconds to sleep in between status checks

        Raises:
            ValueError: if the current state is `init`
        """
        if timeout is None:
            timeout = inf
        if self._run_machine.state == RunMachine.Code("init"):
            raise ValueError("Still in state 'init'! Call run() first!")
        until = RunMachine.Code(until)
        start = time.monotonic()
        while until != self._run_machine.state and time.monotonic() - start < timeout:
            time.sleep(sleep)

    @property
    def status(self):
        return self._run_machine._data["status"]

    @property
    def output(self):
        return self._run_machine._data["output"]


class Executor:
    def submit(self, tasks: List[AbstractTask]) -> ExecutionContext:
        return ExecutionContext(tasks)


from concurrent.futures import (
    ThreadPoolExecutor,
    ProcessPoolExecutor,
)
from threading import Lock


def run_task(task):
    return task.execute()


class FuturesExecutionContext(ExecutionContext):
    def __init__(self, pool, tasks):
        super().__init__(tasks=tasks)
        self._pool = pool
        # self._max_tasks = max_tasks if max_tasks is not None else 4
        self._done = 0
        self._futures = {}
        self._subcontexts = {}
        self._generators = {}
        self._status = {}
        self._output = {}
        self._index = {}
        self._lock = Lock()

    def _process_future(self, future):
        task = self._futures[future]

        # do not specify a timeout even though we already know that the future
        # has finished.  For executors with high latency, low number of workers
        # and large number of tasks transfering the results back to this thread
        # can take longer than one would naively assume (>1ms).  If this is the
        # case we might trip the timeout while waiting for the result, botching
        # the calculation
        status, output = future.result()
        with self._lock:
            self._status[task] = status
            self._output[task] = output
            self._done += 1
        self._check_finish()

    def _prepare_subcontext(self, task, sub_tasks):
        sub = self._subcontexts[task] = type(self)(self._pool, sub_tasks)
        sub._run_machine.observe("finished", partial(self._process_generator, task))
        sub.run()
        return sub

    def _process_generator(self, task, _data):
        gen = self._generators[task]
        sub = self._subcontexts[task]
        try:
            tasks = gen.send(list(zip(sub.status, sub.output)))
            self._prepare_subcontext(task, tasks)
        except StopIteration as stop:
            with self._lock:
                self._status[task], self._output[task] = stop.args[0]
                self._done += 1
                del self._subcontexts[task]
                del self._generators[task]
            self._check_finish()

    def _check_finish(self, log=False):
        with self._lock:
            if self._done == len(self.tasks):
                status = [
                    self._status[n]
                    for n in sorted(self.tasks, key=lambda n: self._index[n])
                ]
                output = [
                    self._output[n]
                    for n in sorted(self.tasks, key=lambda n: self._index[n])
                ]
                self._run_machine.step(
                    "collect",
                    status=status,
                    output=output,
                )

    def _run_running(self):
        if len(self._futures) == 0:
            for i, task in enumerate(self.tasks):
                self._index[task] = i
                if isinstance(task, TaskGenerator):
                    gen = self._generators[task] = iter(task)
                    self._prepare_subcontext(task, next(gen))
                else:
                    future = self._pool.submit(run_task, task)
                    self._futures[future] = task
                    future.add_done_callback(self._process_future)
        else:
            logging.info("Some tasks are still executing!")


class BackgroundExecutor(Executor):
    def __init__(self, max_threads):
        self._max_threads = max_threads
        self._pool = None

    def submit(self, tasks):
        if self._pool is None:
            self._pool = ThreadPoolExecutor(max_workers=self._max_threads)
        return FuturesExecutionContext(self._pool, tasks)


class ProcessExecutor(Executor):
    def __init__(self, max_processes):
        self._max_processes = max_processes
        self._pool = None

    def submit(self, tasks):
        if self._pool is None:
            self._pool = ProcessPoolExecutor(max_workers=self._max_processes)
        return FuturesExecutionContext(self._pool, tasks)


from dask.distributed import Client, LocalCluster


class DaskExecutor(Executor):
    def __init__(self, client):
        self._client = client

    @classmethod
    def from_cluster(cls, cluster):
        return cls(Client(cluster))

    @classmethod
    def from_localcluster(cls, max_processes=None, **kwargs):
        return cls(Client(LocalCluster(n_workers=max_processes, **kwargs)))

    def submit(self, tasks):
        return FuturesExecutionContext(self._client, tasks)


from pympipool import PoolExecutor


class PyMPIExecutor(Executor):
    def __init__(self, max_workers, **kwargs):
        self._max_workers = max_workers
        self._pool = PoolExecutor(max_workers=max_workers, **kwargs)

    def submit(self, tasks):
        return FuturesExecutionContext(self._pool, tasks)



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

from pyiron_contrib.workflow.node import Node
from pyiron_contrib.workflow.util import DotDict

if TYPE_CHECKING:
    from pyiron_contrib.workflow.composite import Composite


class NodePackage(DotDict):
    """
    A collection of node classes that, when instantiated, will have their workflow
    automatically set.

    Node classes are accessible by their _class name_ by item or attribute access.

    Can be extended by adding node classes to new names with an item or attribute set,
    but to update an existing node the `update` method must be used.
    """

    def __init__(self, parent: Composite, *node_classes: Node):
        super().__init__()
        self.__dict__["_parent"] = parent  # Avoid the __setattr__ override
        for node in node_classes:
            self[node.__name__] = node

    def __setitem__(self, key, value):
        if key in self.keys():
            raise KeyError(f"The name {key} is already a stored node class.")
        elif key in self.__dir__():
            raise KeyError(
                f"The name {key} is already an attribute of this "
                f"{self.__class__.__name__} instance."
            )
        if not isinstance(value, type) or not issubclass(value, Node):
            raise TypeError(
                f"Can only set members that are (sub)classes of  {Node.__name__}, "
                f"but got {type(value)}"
            )
        super().__setitem__(key, value)

    def __getitem__(self, item):
        value = super().__getitem__(item)
        if issubclass(value, Node):
            return partial(value, parent=self._parent)
        else:
            return value

    def update(self, *node_classes):
        replacing = set(self.keys()).intersection([n.__name__ for n in node_classes])
        for name in replacing:
            del self[name]

        for node in node_classes:
            self[node.__name__] = node
